import numpy as np
import sys
from .utils import *

def main():

 filename = sys.argv[1]
 if len(sys.argv) == 3:
    output = sys.argv[2]
 else  :
    output = 'rta'


 tmp = np.loadtxt(filename,skiprows=1,delimiter=',')
 nq = int(np.max(tmp[:,0]))+1
 nb = int(np.max(tmp[:,1]))+1
 ntot = nq*nb

 tau = tmp[:,7]
 v   = tmp[:,8:]
 C   = tmp[:,6]
 w   = tmp[:,5]
 kappa =  np.einsum('ki,kj,k,k',v,v,tau,C) 

 data = {'C':C,'tau':tau,'v':v,'kappa':kappa,'f':w/2.0/np.pi}

 save_data(output,data)   



import plotly.offline as py
import plotly
import plotly.graph_objs as go
import numpy as np
import sys
from .utils import *
import time


def plot_results(solver,geometry,**argv):


   data  = solver['variables']
   nodes = geometry['nodes']
   elems = np.array(geometry['elems'])
   size  = [max(nodes[:,i]) - min(nodes[:,i])  for i in range(3)] 
   dim   = 2   if size[2] == 0 else 3


   nn = 0
   nv = len(data.keys())
   #----------------------

   #Create button
   buttons = [dict(label=name,method="update",args=[{"visible":(np.eye(nv) == 1)[k]},])   for k,name in enumerate(data.keys())]


   fig = go.Figure()

   for kk,(name,value) in enumerate(data.items()):

     z = np.zeros(len(nodes))  if np.shape(nodes)[1] == 2 else nodes[:,2]
     fig.add_trace(go.Mesh3d(x=nodes[:,0],
                     y=nodes[:,1],
                     z=z,
                     colorscale='Viridis',
                     colorbar_title = '[' + value['units'] + ']' if len(value['units']) > 0 else "",
                     intensity = value['data'],
                     colorbar=dict(len=0.5),
                     intensitymode='vertex',i=elems[:,0],j=elems[:,1],k=elems[:,2],
                     name=name,
                     showscale = True,
                     visible= True if kk == nn else False
                    ))

   fig.update_layout(
    font=dict(
        family="Courier New, monospace",
        size=12,
        color="black"
    )
   )


   fig.update_layout(
    title={
        'text': "OpenBTE",
        'y':0.93,
        'x':0.88,
        'xanchor': 'center',
        'yanchor': 'top'})


   #if argv.setdefault('large',False):
   #    x1 = 700
   #    x2 = 500
   #else:    
   x1 = 400
   x2 = 450

   fig.update_layout(width=x1,height=x2,autosize=True,margin=dict(t=20, b=10, l=00, r=20),paper_bgcolor='rgba(0,0,0,0)',plot_bgcolor='rgba(0,0,0,0)')  

   updatemenus=[dict(direction='down',active=nn,x=0.4,y=1.15,buttons=list(buttons),showactive=True)]


   #update axes---------------
   axis = dict( 
          backgroundcolor="rgb(255, 255,255)",
          gridcolor="rgb(255, 255, 255)",
          zerolinecolor="rgb(255, 255, 255)",
          visible=False,
          showbackground=True
         )

   #axis = dict(ticktext=[],tickvals= [],showbackground=False)

   bb = str(round(solver['kappa_bte'],2))+' W/m/K' if 'kappa_bte' in solver.keys() else '--'

   meta  = 'Bulk: ' + str(round(solver['kappa_bulk'],2)) +' W/m/K<br>Fourier: '\
                    +       str(round(solver['kappa_fourier'],2)) + ' W/m/K<br>BTE:' \
                    +       bb



   fig.add_annotation(
            x=0.97,
            y=0.05,
            xref='paper',
            showarrow=False,
            yref='paper',
            text=meta,align='left')


   #fig.add_annotation(annotations)

   fig.update_layout(showlegend=False)

   # Update 3D scene options
   fig.update_scenes(
    aspectratio=dict(x=1, \
                     y=size[1]/size[0], \
                     z= 1 if dim == 2 else size[2]/size[0]),
    aspectmode="manual",
    xaxis=dict(axis),
    yaxis=dict(axis),
    zaxis=dict(axis))

   #camera
   camera = dict(
    up=dict(x=0, y=1, z=0),
    center=dict(x=0, y=0, z=0),
    eye=dict(x=0, y=0, z=2)
   )

   fig.update_layout(xaxis_showgrid=False, yaxis_showgrid=False)
   fig.update_layout(scene_camera=camera)
   fig.update_layout(updatemenus=updatemenus)

   if argv.setdefault('write_html',False):
    fig.write_html("plotly.html")
   #plotly.io.write_json(fig,'ff.json')

   # plotly.io.to_image(fig,format='png')
   #fig.write_image("test.png",scale=5)
   if 'google.colab' in sys.modules:
     fig.show(renderer='colab')
   else:
     fig.show(renderer='browser')


   return fig




from openbte import Material,Solver,Geometry,Plot
import numpy as np
import yaml
import re,os

def main():

 if len(os.sys.argv) == 1:
   name = 'input.yaml'  
   with open(name) as f:
    data = yaml.safe_load(f)
 else:   
  a = os.sys.argv[1]
  if ':' in a:
   data = yaml.safe_load(a)

  else: 
   name = os.sys.argv[1]   
   with open(name) as f:
    data = yaml.safe_load(f)

 if 'Material' in data.keys():
  Material(**data['Material'])

 if 'Geometry' in data.keys():
  Geometry(**data['Geometry'])

 if 'Solver' in data.keys():
  Solver(**data['Solver'])

 if 'Plot' in data.keys():
  Plot(**data['Plot'])

 




import numpy as np
from openbte.utils import *
from pubeasy import MakeFigure



def plot_suppression(**argv):


 sol = load_data('solver')
 T_sampled = sol['kappa_mode'].T

 if argv.setdefault('fourier',False):
  T_sampled_f = sol['kappa_mode_f'].T

 if argv.setdefault('zero',False):
  T_sampled_0 = sol['kappa_0'].T


 #----------------------------------
 #Compute MFPs
 mat = load_data('material')
 mfp_sampled  = mat['mfp_sampled']*1e9
 sigma  = mat['sigma']*1e9

 vmfp  = mat['VMFP']
 [n_phi,n_theta,n_mfp]  = mat['sampling']

 Dphi = 2*np.pi/n_phi
 phi = np.linspace(Dphi/2,2.0*np.pi-Dphi/2,n_phi,endpoint=True)
 phi = list(phi)
 bmfp = np.linalg.norm(vmfp,axis=1)
 mfp_vec = np.outer(vmfp[:,0],mfp_sampled)
 M = mfp_vec
 #-----------------------------

 S = T_sampled/M*1e18
 if argv.setdefault('fourier',False):
  Sf = T_sampled_f/M*1e18
  Smf = np.mean(Sf,axis=0)
 Sm = np.mean(S,axis=0)
 
 if argv.setdefault('show',True):
  fig = MakeFigure()
  fig.add_plot(mfp_sampled*1e-3,Sm,color='k',name='bte')
  if argv.setdefault('fourier',False):
   fig.add_plot(mfp_sampled*1e-3,Smf,color='g',name='fourier')
  fig.add_labels('Mean Free Path [$\mu$ m]','Suppression')
 fig.finalize(grid=True,xscale='log',write = False,show=True,ylim=[0,1])

 if argv.setdefault('mode_resolved',False):
 
     nm = n_mfp
     phis = range(n_phi)

     for p in phis:
         fig = MakeFigure()
         fig.add_plot(mfp_sampled[:nm]*1e-3,S[p,:nm],color='k',marker='o')
         fig.add_plot(mfp_sampled[:nm]*1e-3,Sf[p,:nm],color='r',marker='o')
         fig.finalize(grid=True,xscale='log',write = False,show=True)
    

 save_data(argv.setdefault('suppression_file','suppression'),{'mfp':mfp_sampled,'suppression':Sm})

 return mfp_sampled,Sm

if __name__ == '__main__':

    main(show=True)



from shapely.ops import cascaded_union
from shapely.geometry import Point,LineString
from shapely.geometry import MultiPolygon
from shapely.geometry import Polygon
import shapely
import numpy as np
from .shape import *
from .utils import *
import subprocess
import os
import time

class Mesher(object):

 def __init__(self,argv):

  argv['dmin'] = 0
  model = argv.setdefault('model','lattice')

  if model == 'lattice':   
   #create polygons-----
   self.add_symmetry(argv) 
   shapes = get_shape(argv)

   polygons = [translate_shape(shapes[n],i) for n,i in enumerate(argv['base'])]
   argv.update({'polygons':np.array(polygons,dtype=object)})
   repeat_merge_scale(argv)

  elif model == 'custom':
    repeat_merge_scale(argv)

  elif model =='bulk':

    if argv.setdefault('lz',0) == 0:
     return self.generate_bulk_2D(argv)
    else :
     return self.generate_bulk_3D(argv)

  elif model == 'random':
      while argv['dmin'] < argv.setdefault('dmin_prescribed',0.02):  
       argv['base'] = np.random.rand(argv.setdefault('Np',5),2)-0.5 

       self.add_symmetry(argv) 
       self.add_polygons(argv)
       repeat_merge_scale(argv)
  #---------------------

  if argv.setdefault('lz',0) == 0:
    self.generate_mesh_2D(argv)
  else:   
    self.generate_mesh_3D(argv)



 def generate_bulk_3D(self,argv):


  Lx = argv['lx']
  Ly = argv['ly']
  Lz = argv['lz']

  frame = generate_frame(**argv)

  mesh_ext = argv['step']
  Nx = float(argv.setdefault('Nx',1))
  Ny = float(argv.setdefault('Ny',1))
  Nz = float(argv.setdefault('Nz',1))

  Lx *=Nx
  Ly *=Ny
  Lz *=Nz


  store = open('mesh.geo', 'w+')

  points  = [[-Lx/2,-Ly/2,-Lz/2],\
             [-Lx/2,+Ly/2,-Lz/2],\
             [Lx/2,Ly/2,-Lz/2],\
             [Lx/2,-Ly/2,-Lz/2],\
             [-Lx/2,-Ly/2,Lz/2],\
             [-Lx/2,+Ly/2,Lz/2],\
             [Lx/2,Ly/2,Lz/2],\
             [Lx/2,-Ly/2,Lz/2]]

  #Write points
  for p,point in enumerate(points):
   store.write( 'Point('+str(p) +') = {' + str(point[0]) +','+\
                                          str(point[1])+',' +\
                                          str(point[2]) + ','+\
                                          str(mesh_ext) +'};\n')

  #Write lines
  ll = 1
  store.write( 'Line('+str(ll) +') = {' + str(0) +','+ str(1)+'};\n')
  ll += 1
  store.write( 'Line('+str(ll) +') = {' + str(1) +','+ str(2)+'};\n')
  ll += 1
  store.write( 'Line('+str(ll) +') = {' + str(2) +','+ str(3)+'};\n')
  ll += 1
  store.write( 'Line('+str(ll) +') = {' + str(3) +','+ str(0)+'};\n')
  ll += 1
  store.write( 'Line('+str(ll) +') = {' + str(4) +','+ str(5)+'};\n')
  ll += 1
  store.write( 'Line('+str(ll) +') = {' + str(5) +','+ str(6)+'};\n')
  ll += 1
  store.write( 'Line('+str(ll) +') = {' + str(6) +','+ str(7)+'};\n')
  ll += 1
  store.write( 'Line('+str(ll) +') = {' + str(7) +','+ str(4)+'};\n')
  ll += 1
  store.write( 'Line('+str(ll) +') = {' + str(0) +','+ str(4)+'};\n')
  ll += 1
  store.write( 'Line('+str(ll) +') = {' + str(1) +','+ str(5)+'};\n')
  ll += 1
  store.write( 'Line('+str(ll) +') = {' + str(2) +','+ str(6)+'};\n')
  ll += 1
  store.write( 'Line('+str(ll) +') = {' + str(3) +','+ str(7)+'};\n')
  ll += 1

  #Write Surfaces
  loops = [[1,2,3,4],\
           [5,6,7,8],\
           [9,5,-10,-1],\
           [10,6,-11,-2],\
           [12,-7,-11,3],\
           [4,9,-8,-12]]

  for nloop,loop in enumerate(loops) :
   index = nloop + 12
   strc = 'Line Loop(' + str(index)+ ') = {'
   for n,p in enumerate(loop) :
    strc +=str(p)
    if n == len(loop)-1:
     strc += '};\n'
    else :
     strc += ','
   store.write(strc)
   strc = 'Plane Surface(' + str(nloop+1)+ ') = {'+ str(index) + '};\n'
   store.write(strc)

  strc = 'Surface Loop(' + str(12)+ ') = {-1,3,4,5,6,2};\n'
  store.write(strc)
  strc = 'Volume(0)={12};\n'
  store.write(strc)
  strc = 'Physical Volume("Matrix")={0};\n'
  store.write(strc)



  #Write Periodic Surfaces--------------------------------------------
  strc = 'Periodic Surface 1 {1,2,3,4} = 2 {5,6,7,8};\n'
  store.write(strc)
  strc = 'Periodic Surface 3 {9,5,-10,-1} = 5 {12,-7,-11,3};\n'
  store.write(strc)
  strc = 'Periodic Surface 4 {10,6,-11,-2} = 6 {9,-8,-12,4};\n'
  store.write(strc)
  #------------------------------------------------------------------


  boundary_surfaces = []
  if argv.setdefault('Periodic',[True,True,True])[0]:
   strc = r'''Physical Surface('Periodic_1') = {5} ;''' + '\n'
   store.write(strc)
   strc = r'''Physical Surface('Periodic_2') = {3} ;''' + '\n'
   store.write(strc)
  else:
   boundary_surfaces.append(3)
   boundary_surfaces.append(5)

  if argv.setdefault('Periodic',[True,True,True])[1]:
   strc = r'''Physical Surface('Periodic_3') = {4} ;''' + '\n'
   store.write(strc)
   strc = r'''Physical Surface('Periodic_4') = {6} ;''' + '\n'
   store.write(strc)
  else:
   boundary_surfaces.append(4)
   boundary_surfaces.append(6)

  if argv.setdefault('Periodic',[True,True,True])[2]:
   strc = r'''Physical Surface('Periodic_5') = {2} ;''' + '\n'
   store.write(strc)
   strc = r'''Physical Surface('Periodic_6') = {1} ;''' +'\n'
   store.write(strc)
  else:
   boundary_surfaces.append(1)
   boundary_surfaces.append(2)


  #write Boundary surfaces
  if len(boundary_surfaces) > 0:
   strc = r'''Physical Surface('Boundary') = {'''
   for p,side in enumerate(boundary_surfaces) :
    strc +=str(side)
    if p == len(boundary_surfaces)-1:
     strc += '};\n'
    else :
     strc += ','
   store.write(strc)


  store.close()
  with open(os.devnull, 'w') as devnull:
    output = subprocess.check_output("gmsh -optimize_netgen -format msh2 -3 mesh.geo -o mesh.msh".split(), stderr=devnull)


 def add_polygons(self,argv):

   shapes = get_shape(argv)
   polygons = [translate_shape(shapes[n],i) for n,i in enumerate(argv['base'])]
   argv.update({'polygons':np.array(polygons)})


 def add_symmetry(self,argv):

  if argv.setdefault('invariance',False):
      argv['base'] =  np.append(argv['base'],[[0,0]],axis=0)

  if argv.setdefault('reflection',False):
     
      argv['base']=np.array(argv['base'])
      base = argv['base'].copy()
      tmp = base/2 + np.array([0.25,0.25])
      base = argv['base'].copy()
      tmp =  np.append(tmp,-base/2 - [0.25,0.25],axis=0)
      base = argv['base'].copy()
      base[:,1] = -base[:,1]
      tmp =  np.append(tmp,base/2 - [-0.25,0.25],axis=0)
      base = argv['base'].copy()
      base[:,0] = -base[:,0]
      tmp =  np.append(tmp,base/2 - [0.25,-0.25],axis=0)
      argv['base'] = tmp



 def generate_bulk_2D(self,argv):

  direction = argv.setdefault('direction','x')

  frame = generate_frame(**argv)
  Lx = float(argv['lx'])
  Ly = float(argv['ly'])

  mesh_ext = argv['step']
  points = []
  lines = []

  store = open("mesh.geo", 'w+')
  store.write('h='+str(mesh_ext) + ';\n')

  for k,p in enumerate(frame) :
   store.write( 'Point('+str(k) +') = {' + str(p[0]) +','+ str(p[1])+',0,h};\n')

  ll = 0
  for k,p in enumerate(frame) :
   p1 = k
   p2 = (k+1)%len(frame)
   ll += 1
   store.write( 'Line('+str(ll) +') = {' + str(p1) +','+ str(p2)+'};\n')

  strc = 'Line Loop(5) = {'
  for p,point in enumerate(frame) :
   strc +=str(p+1)
   if p == len(frame)-1:
    strc += '};\n'
   else :
    strc += ','
  store.write(strc)

  #Create Surface
  strc = 'Plane Surface(10) = {5};\n'
  store.write(strc)
  strc = r'''Physical Surface('Matrix') = {10};'''+'\n'
  store.write(strc)

  bs = []
  if argv.setdefault('Periodic',[True,True,True])[1] :
    strc = r'''Physical Line('Periodic_1') = {1};''' + '\n'
    store.write(strc)
    strc = r'''Physical Line('Periodic_2') = {3};''' + '\n'
    store.write(strc)
  else:
   if direction=='y':
    strc = r'''Physical Line('Cold') = {2};''' + '\n'
    store.write(strc)
    strc = r'''Physical Line('Hot') = {4};''' +'\n'
    store.write(strc)
   else:
    bs.append(1)
    bs.append(3)

  if argv.setdefault('Periodic',[True,True,True])[0] :
    strc = r'''Physical Line('Periodic_3') = {2};''' + '\n'
    store.write(strc)
    strc = r'''Physical Line('Periodic_4') = {4};''' +'\n'
    store.write(strc)
  else:
   if direction=='x':
    strc = r'''Physical Line('Hot') = {2};''' + '\n'
    store.write(strc)
    strc = r'''Physical Line('Cold') = {4};''' +'\n'
    store.write(strc)
   else:
    bs.append(2)
    bs.append(4)


  if len(bs) > 0:
   strc = r'''Physical Line('Boundary') = {'''
   for p,side in enumerate(bs) :
    strc +=str(side)
    if p == len(bs)-1:
     strc += '};\n'
    else :
     strc += ','
   store.write(strc)


  strc = 'Periodic Line{1}={-3};\n'
  store.write(strc)
  strc = 'Periodic Line{2}={-4};\n'
  store.write(strc)

#-------------------------------------------------------
  store.close()

  
  with open(os.devnull, 'w') as devnull:
    output = subprocess.check_output("gmsh -format msh2 -2 mesh.geo -o mesh.msh".split(), stderr=devnull)




 def create_loop(self,loops,line_list,store):

   #create external loop
   strc = 'Line Loop(' + str(loops)+ ') = {'
   for n in range(len(line_list)):
    strc +=str(line_list[n])
    if n == len(line_list)-1:
     strc += '};\n'
    else:
     strc += ','
   store.write(strc)

   return strc

 def generate_mesh_2D(self,argv): 

  polygons = argv['polygons']
  mesh_ext = argv['step']
 
  #Create the frame
  frame = generate_frame(**argv)
  Frame = Polygon(frame)

  #CREATE BULK SURFACE------------------------------------
  store = open('mesh.geo', 'w+')
  store.write('h='+str(mesh_ext) + ';\n')
  store.write('lx='+str(argv['lx']) + ';\n')
  store.write('ly='+str(argv['ly']) + ';\n')
  points = []
  lines = []
  loops = 1000
  ss = 0

  polypores = []
  for poly in polygons:
   polypores.append(Polygon(poly))

  #Get boundary wall
  self.lx = abs(frame[0][0])*2.0
  self.ly = abs(frame[0][1])*2.0
  pore_wall = []
  delta = 1e-2
  #-------------------------


  #for pore in polygons:
  #    for p in pore:
  #      plt.scatter(p[0],p[1])  
  #plt.show()

  
  bulk = Frame.difference(cascaded_union(polypores))
  if not (isinstance(bulk, shapely.geometry.multipolygon.MultiPolygon)):
   bulk = [bulk]

  bulk_surface = []
  inclusions = []
 
  for r,region in enumerate(bulk):
   
    pp = list(region.exterior.coords)[:-1]
    line_list = self.create_line_list(pp,points,lines,store)

    #for l in lines:
    #  p1 = points[l[0]]
    #  p2 = points[l[1]]
    #  plt.plot([p1[0],p2[0]],[p1[1],p2[1]])
    #plt.show()
    #quit()
    loops +=1
    local_loops = [loops]
    self.create_loop(loops,line_list,store)

    #Create internal loops-------------
    a = 0
    for interior in region.interiors:
     a +=1
     pp = list(interior.coords)[:-1]
     line_list = self.create_line_list(pp,points,lines,store)
     loops +=1
     local_loops.append(loops)
     self.create_loop(loops,line_list,store)
    #---------------------------------
    ss +=1
    bulk_surface.append(ss)
    self.create_surface_old(local_loops,ss,store)

  #Create Inclusion surfaces--------
  if argv.setdefault('inclusion',False):

    for poly in polygons:
     thin = Polygon(poly).intersection(Frame)

     #Points-------
     pp = list(thin.exterior.coords)[:-1]
     line_list = create_line_list(pp,points,lines,store)
     loops +=1
     self.create_loop(loops,line_list,store)
     ss +=1
     create_surface([loops],ss,store)
     inclusions.append(ss)

    strc = r'''Physical Surface('Inclusion') = {'''
    for n,s in enumerate(inclusions):
     strc += str(s)
     if n == len(inclusions)-1:
      strc += '};\n'
     else:
       strc += ','
    store.write(strc)
  #-------------------------------


  strc = r'''Physical Surface('Matrix') = {'''
  for n,s in enumerate(bulk_surface):
   strc += str(s)
   if n == len(bulk_surface)-1:
     strc += '};\n'
   else:
     strc += ','
  store.write(strc)


  hot = []
  cold = []
  upper = []
  lower = []

  pul = np.array([-self.lx/2.0,self.ly/2.0])
  pur = np.array([self.lx/2.0,self.ly/2.0])
  pll = np.array([-self.lx/2.0,-self.ly/2.0])
  plr = np.array([self.lx/2.0,-self.ly/2.0])


  delta = 1e-12
  pore_wall = []
  for l,line in enumerate(lines):
 
   pl = (np.array(points[line[0]])+np.array(points[line[1]]))/2.0
   
   is_on_boundary = True
   if self.compute_line_point_distance(pul,pur,pl) < delta:     
     upper.append(l+1)
     is_on_boundary = False
     
   if self.compute_line_point_distance(pll,plr,pl) < delta:
     lower.append(l+1)
     is_on_boundary = False 
     
   if self.compute_line_point_distance(plr,pur,pl) < delta:
     cold.append(l+1)
     is_on_boundary = False 
     
   if self.compute_line_point_distance(pll,pul,pl) < delta:
     hot.append(l+1)
     is_on_boundary = False   
     

   if is_on_boundary:
    pore_wall.append(l+1)

  additional_boundary = []
  argv.setdefault('Periodic',[True,True,True])
  if argv['Periodic'][0]:
   strc = r'''Physical Line('Periodic_2') = {'''
   for n in range(len(hot)-1):
    strc += str(hot[n]) + ','
   strc += str(hot[-1]) + '};\n'
   store.write(strc)

   strc = r'''Physical Line('Periodic_1') = {'''
   for n in range(len(cold)-1):
    strc += str(cold[n]) + ','
   strc += str(cold[-1]) + '};\n'
   store.write(strc)
  else:
   for k in hot:
    additional_boundary.append(k)
   for k in cold:
    additional_boundary.append(k)

  if argv.setdefault('Periodic',[True,True,True])[1] and len(upper) > 0:
   strc = r'''Physical Line('Periodic_3') = {'''
   for n in range(len(upper)-1):
    strc += str(upper[n]) + ','
   strc += str(upper[-1]) + '};\n'
   store.write(strc)

   strc = r'''Physical Line('Periodic_4') = {'''
   for n in range(len(lower)-1):
    strc += str(lower[n]) + ','
   strc += str(lower[-1]) + '};\n'
   store.write(strc)
  else:
   for k in lower:
    additional_boundary.append(k)
   for k in upper:
    additional_boundary.append(k)

  #Collect Wall
  if argv.setdefault('inclusion',False):
   interface_surfaces = pore_wall
   boundary_surfaces = additional_boundary
  else:
   boundary_surfaces = pore_wall + additional_boundary
   interface_surfaces = []

  #Interface surface
  if len(boundary_surfaces)> 0:
   strc = r'''Physical Line('Boundary') = {'''
   for r,region in enumerate(boundary_surfaces) :
    strc += str(region)
    if r == len(boundary_surfaces)-1:
     strc += '};\n'
     store.write(strc)
    else :
     strc += ','


  if len(interface_surfaces)> 0:
   strc = r'''Physical Line('Interface') = {'''
   for r,region in enumerate(interface_surfaces) :
    strc += str(region)
    if r == len(interface_surfaces)-1:
     strc += '};\n'
     store.write(strc)
    else :
     strc += ','

  if argv['Periodic'][0]:
   for n in range(len(hot)):
    periodic = self.FindPeriodic(hot[n],cold,points,lines) 
    strc = 'Periodic Line{' + str(hot[n]) + '}={' + str(periodic) + '};\n'
    store.write(strc)

  if argv['Periodic'][1]:
   for n in range(len(lower)):
    periodic = self.FindPeriodic(lower[n],upper,points,lines)
    strc = 'Periodic Line{' + str(lower[n]) + '}={' + str(periodic) + '};\n'
    store.write(strc)

#-------------------------------------------------------
  store.close()

  with open(os.devnull, 'w') as devnull:
    output = subprocess.check_output("gmsh -format msh2 -2 mesh.geo -o mesh.msh".split(), stderr=devnull)


 def FindPeriodic(self,control,others,points,lines):
  
  delta = 1e-3
  p1c = np.array(points[lines[control-1][0]])
  p2c = np.array(points[lines[control-1][1]])
  pc = (p1c + p2c)/2.0
 
  for n in range(len(others)):
   p1o = np.array(points[lines[others[n]-1][0]])
   p2o = np.array(points[lines[others[n]-1][1]])
   po = (p1o + p2o)/2.0   
  
  
   if np.linalg.norm(np.dot(p1c-p2c,pc-po)) < delta:     
    line1 = LineString([p1c,p1o])
    line2 = LineString([p2c,p2o])     
    if line1.intersects(line2):
     return -others[n]
    else:
     return others[n]

  #If no periodic line has been found
  print("NO PERIODIC LINES HAS BEEN FOUND")
  quit()


 def create_surface_old(self,ss,bulk_surface,store):

  strc = '''Plane Surface(''' + str(bulk_surface) + ')= {'
  for n,s in enumerate(ss):
   strc += str(s)
   if n == len(ss)-1:
     strc += '};\n'
   else:
     strc += ','
  store.write(strc)


 def compute_line_point_distance(self,p1,p2,p3):
   return np.linalg.norm(np.cross(p2-p1, p3-p1))/np.linalg.norm(p2-p1)


 def generate_mesh_3D(self,argv): 

  self.step = argv['step']
  self.points = []
  self.lines = []
  self.loops = []
  self.surfaces = []
  self.argv = argv
  self.common(z=-self.argv['lz']/2)

  self.common(z=self.argv['lz']/2)

  self.merge() 
  self.points = np.array(self.points)
  self.lines = np.array(self.lines)
  self.loops = np.array(self.loops)
  
  self.apply_periodic_mesh()
 
  self.write_geo()

     
 def get_surface_centroid(self,ss):
     
  
   s = self.surfaces[ss][0]  
    
   pts = []
   for n in self.loops[s]:
    if n > 0:
        i = 0
    else:
        i = 1
    pts.append(self.points[self.lines[abs(n)-1][i]])
   pts = np.array(pts) 
   
   return np.mean(pts,axis=0)
   
     

 def isperiodic(self,s1,s2):

   c1 = self.get_surface_centroid(s1)
   c2 = self.get_surface_centroid(s2)
   
   
   if abs(np.linalg.norm(c1-c2) - self.argv['lx'])<1e-4 and \
      np.linalg.norm(np.cross(c1-c2,[1,0,0])) < 1e-4: 
      return 0,c1-c2
   

   if abs(np.linalg.norm(c1-c2) - self.argv['ly'])<1e-4 and \
      np.linalg.norm(np.cross(c1-c2,[0,1,0])) < 1e-4: 
      return 1,c1-c2
   
   if abs(np.linalg.norm(c1-c2) - self.argv['lz'])<1e-4 and \
      np.linalg.norm(np.cross(c1-c2,[0,0,1])) < 1e-4: 
      return 2,c1-c2
   
  
    
   return None,None
   
 
   
    
 def get_points_from_surface(self,s1):
      
   ll = self.surfaces[s1][0]
   pts = []
   ind = []
   for n in self.loops[ll]:
    if n > 0:
        i = 0
    else:
        i = 1
    pts.append(self.points[self.lines[abs(n)-1][i]])
    
    ind.append(self.lines[abs(n)-1][i])

   return np.array(pts),ind
     
     
 def apply_periodic_mesh(self):

     self.argv.setdefault('Periodic',[True,True,False])
     #Find periodic surfaces----
     self.periodic = {}
     for s1 in range(len(self.surfaces)):
      for s2 in range(s1+1,len(self.surfaces)):
         (a,per) = self.isperiodic(s1,s2)
         if not a == None and self.argv['Periodic'][a]:
           
           corr = {}
           pts1,ind1 = self.get_points_from_surface(s1)
           pts2,ind2 = self.get_points_from_surface(s2)

           for n1,p1 in zip(ind1,pts1):
            for n2,p2 in zip(ind2,pts2):
             if np.linalg.norm(p1-p2-per) < 1e-3:
                corr[n1] = n2
                break            
           #-----------------------------------------
           loop_1 = self.loops[self.surfaces[s1][0]] 
           loop_2 = []
           for ll in loop_1 :
             pts = self.lines[abs(ll)-1]
             if ll > 0:
              p1 = pts[0]; p2 = pts[1] 
             else:
              p1 = pts[1]; p2 = pts[0]   
             loop_2.append(self.line_exists_ordered(corr[p1],corr[p2]))             
           self.periodic[s1] = [s2,loop_1,loop_2,np.array(per/np.linalg.norm(per))]  
           
           break  
     
      
        
       

 def merge(self):

   #Create lines
   Nh = int(len(self.points)/2)
   for n in range(Nh):
     self.lines.append([n,n+Nh])  

   Ns = int(len(self.surfaces)/2)
   self.surfaces3D = []
   for s in range(Ns):
    tmp2 = [s,s+Ns]  

    for loop in self.surfaces[s]:

     #Create vertical loop    
     for l in self.loops[loop]:
      i1 = self.lines[l-1][0];  i2 = self.lines[l-1][1]
      j1 = i1 + Nh; j2 = i2 + Nh
      tmp = []
      tmp.append(self.line_exists_ordered(i1,i2))
      tmp.append(self.line_exists_ordered(i2,j2))
      tmp.append(self.line_exists_ordered(j2,j1))
      tmp.append(self.line_exists_ordered(j1,i1))
      self.loops.append(tmp)
      self.surfaces.append([len(self.loops)-1])
      tmp2.append(len(self.surfaces)-1)

 
    self.surfaces3D.append(tmp2)

   '''
   #Create new loops 
   N = int(len(self.loops)/2)
   for ll in range(N):
    for l in self.loops[ll]:         
     
     i1 = self.lines[l-1][0];  i2 = self.lines[l-1][1]
              
     j1 = i1 + Nh; j2 = i2 + Nh
     tmp = []
     tmp.append(self.line_exists_ordered(i1,i2))
     tmp.append(self.line_exists_ordered(i2,j2))   
     tmp.append(self.line_exists_ordered(j2,j1))
     tmp.append(self.line_exists_ordered(j1,i1))
  
     self.loops.append(tmp)
     self.surfaces.append([len(self.loops)-1])
   print(self.loops)
   '''   
 
 #def apply_periodic_mesh(self):

 def write_geo(self):

   
   store = open('mesh.geo', 'w+')

   
   #Write points-----
   for p,point in enumerate(self.points):
    store.write( 'Point('+str(p) +') = {' + str(point[0]) +','+ str(point[1])+',' + str(point[2])+','+ str(self.step) +'};\n')
   
   #Write lines----
   for l,line in enumerate(self.lines):
     store.write( 'Line('+str(l+1) +') = {' + str(line[0]) +','+ str(line[1])+'};\n')
      
   #Write loops------
   nl = len(self.lines)+1
   for ll,loop in enumerate(self.loops):  
    strc = 'Line Loop(' + str(ll+nl)+ ') = {'
    for n,l in enumerate(loop):
     strc +=str(l)
     if n == len(loop)-1:
      strc += '};\n'
     else:
      strc += ','
    store.write(strc)
   
   
   #Write surface-----
   nll = len(self.loops)
   for s,surface in enumerate(self.surfaces):
    strc = '''Plane Surface(''' +str(s+nll+nl) + ''')= {''' 
    for n,ll in enumerate(surface):
      strc += str(ll+nl)  
      if n == len(surface)-1:
       strc += '};\n'
      else:
       strc += ','    
    store.write(strc)
    

   boundary = list(range(len(self.surfaces)))
   #Apply periodicity
   vec = {}

   for key, value in self.periodic.items():
   
     boundary.remove(key)  
     boundary.remove(value[0])  

     value[3] = np.round(value[3],4)     
     store.write('Periodic Surface ' + str(key+nl+nll) + ' {') 
     
     for n,k in enumerate(value[1]):
      store.write(str(k))
      if n < len(value[1])-1:
       store.write(',')
     store.write('} = ' + str(value[0]+nl+nll) + ' {')
     
     for n,k in enumerate(value[2]):
      store.write(str(k))
      if n < len(value[2])-1:
       store.write(',')
     store.write('};\n')
    
     value[0] += nl+nll
     tt = key +  nl + nll
     if np.array_equal(np.array([1,0,0]),value[3]) : 
        vec.setdefault(2,[]).append(value[0]) 
        vec.setdefault(1,[]).append(tt) 
     
     if np.array_equal(np.array([-1,0,0]),value[3]): 
        vec.setdefault(1,[]).append(value[0]) 
        vec.setdefault(2,[]).append(tt) 
        
     if np.array_equal(np.array([0,1,0]),value[3]) : 
        vec.setdefault(3,[]).append(tt) 
        vec.setdefault(4,[]).append(value[0])
        
     if np.array_equal(np.array([0,-1,0]),value[3]) : 
        vec.setdefault(4,[]).append(tt) 
        vec.setdefault(3,[]).append(value[0])
        

     if np.array_equal(np.array([0,0,1]),value[3]) : 
        vec.setdefault(5,[]).append(value[0]) 
        vec.setdefault(6,[]).append(tt)   
        
     if np.array_equal(np.array([0,0,-1]),value[3]) : 
        vec.setdefault(6,[]).append(value[0]) 
        vec.setdefault(5,[]).append(tt) 
   for key, value in vec.items():
     strc = r'''Physical Surface("Periodic_''' + str(key) + '''") = {'''
     for n,p in enumerate(value) :
      strc +=str(p)
      if n == len(value)-1:
       strc += '};\n'
      else :
       strc += ','
     store.write(strc)

   strc = r'''Physical Surface("Boundary") = {'''
   for n,p in enumerate(boundary) :
      strc +=str(p+nl+nll)
      if n == len(boundary)-1:
       strc += '};\n'
      else :
       strc += ','
   store.write(strc)
    
  
   #this part needs to change---

   index = nll+nl +  len(self.surfaces3D)
   for k,surfaces in enumerate(self.surfaces3D):
       
    #------------------
    strc = r'''Surface Loop(''' + str(nll+nl + k) + ') = {'
    for s,surface in enumerate(surfaces) :
     strc +=str(nll+nl+surface)
     if s == len(surfaces)-1:
      strc += '};\n'
     else :
      strc += ','
    store.write(strc)
    
    strc = r'''Volume(''' + str(index + k) + ''') = {''' + str(nll+nl+k) + '};\n'
    store.write(strc)
  
   strc = r'''Physical Volume('Bulk') = {'''
   for k in range(len(self.surfaces3D)) :
     strc +=str(k+index)
     if k == len(self.surfaces3D)-1:
      strc += '};\n'
     else :
      strc += ','
   store.write(strc)

   #---------------------------------------------
    
   store.close()


   with open(os.devnull, 'w') as devnull:
    output = subprocess.check_output("gmsh -optimize_netgen -format msh2 -3 mesh.geo -o mesh.msh".split(), stderr=devnull)
    #output = subprocess.check_output("gmsh -format msh2 -3 mesh.geo -o mesh.msh".split(), stderr=devnull)


 def line_exists_ordered(self,p1,p2):
     
  l = [p1,p2]   
  for n,line in enumerate(self.lines) :
   if (line[0] == l[0] and line[1] == l[1]) :
    return n+1
   if (line[0] == l[1] and line[1] == l[0]) :
    return -(n+1)
  return 0

 def already_included(self,p):

  for n,point in enumerate(self.points):
   d = np.linalg.norm(np.array(point)-np.array(p))
   if d < 1e-12:
    return n
  return -1


 def already_included_old(self,all_points,new_point,p_list):


  dd = 1e-12
  I = []
  if len(all_points) > 0:
   I = np.where(np.linalg.norm(np.array(new_point)-np.array(all_points),axis=1)<dd)[0]
  if len(I) > 0: 
     return I[0]
  else: 
     return -1



 # return loop
 def line_exists_ordered_old(self,l,lines):
  for n,line in enumerate(lines) :
   if (line[0] == l[0] and line[1] == l[1]) :
    return n+1
   if (line[0] == l[1] and line[1] == l[0]) :
    return -(n+1)
  return 0

 def create_line_list(self,pp,points,lines,store):

   #Eliminate unncessesary points--


   #------------------------------

   p_list = []
   for p in pp:
    tmp = self.already_included_old(points,p,p_list)
    if tmp == -1:    
      points.append(p)
      p_list.append(len(points)-1)
    else:
      p_list.append(tmp)

   for k,p in enumerate(p_list):
     point = points[-len(p_list)+k] 
     store.write( 'Point('+str(len(points)-len(p_list)+k) +') = {' + str(point[0]/self.lx) +'*lx,'+ str(point[1]/self.ly)+'*ly,0,h};\n')

   

   line_list = []
   for l in range(len(p_list)):
    p1 = p_list[l]
    p2 = p_list[(l+1)%len(p_list)]
    if not p1 == p2:
    #f 1 == 1:    
     tmp = self.line_exists_ordered_old([p1,p2],lines)

     if tmp == 0 : #craete line
      lines.append([p1,p2])
      store.write( 'Line('+str(len(lines)) +') = {' + str(p1) +','+ str(p2)+'};\n')
      line_list.append(len(lines))
     else:
      line_list.append(tmp)

   
   return line_list


 


 def create_lines_and_loop_from_points(self,pp,z):
     
   p_list = []  
   for p in pp:
    p = [p[0],p[1],z]   
    tmp = self.already_included(p) 
    if tmp == -1:
      self.points.append(p)
      p_list.append(len(self.points)-1)
    else:
      p_list.append(tmp)

   line_list = []
   for l in range(len(p_list)):
    p1 = p_list[l]
    p2 = p_list[(l+1)%len(p_list)]
    if not p1 == p2:
     tmp = self.line_exists_ordered(p1,p2)
     if tmp == 0 : #craete line
      self.lines.append([p1,p2])
      line_list.append(len(self.lines))
     else:
      line_list.append(tmp)
      
   self.loops.append(line_list)  
      

 def common(self,z):

  Frame = Polygon(generate_frame(**self.argv))
  polygons = self.argv['polygons']

  polypores = [ Polygon(poly)  for poly in polygons]
  
  MP = MultiPolygon(polypores)
  bulk = Frame.difference(cascaded_union(MP))
  if not (isinstance(bulk, shapely.geometry.multipolygon.MultiPolygon)):
   bulk = [bulk]

  for region in bulk:
   loop_start = len(self.loops)  
   pp = list(region.exterior.coords)[:-1]
   self.create_lines_and_loop_from_points(pp,z) 
   #From pores 
   for interior in region.interiors:
    pp = list(interior.coords)[:-1]
    self.create_lines_and_loop_from_points(pp,z)
   loop_end = len(self.loops) 

   self.surfaces.append(range(loop_start,loop_end))

    







 

  







import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)
import numpy as np
from .mesher import *
from .utils import *
import time
import scipy.sparse as sp
import itertools
from mpi4py import MPI
import time
import numpy.testing as npt
from statistics import mean
from scipy.ndimage.interpolation import shift
from collections import Counter

comm = MPI.COMM_WORLD

def compute_boundary_condition_data(data,**argv):

    
    side_elem_map = data['side_elem_map_vec']
    face_normals = data['face_normals']
    volumes = data['volumes']
    dim = data['dim']

    direction = data['direction']
    if direction == 0:
     gradir = 0
     applied_grad = [1,0,0]
    if direction == 1:
     gradir = 1
     applied_grad = [0,1,0]
    if direction == 2:
     applied_grad = [0,0,1]
     gradir = 2

    if gradir == 0:
     flux_dir = [1,0,0]
     length = data['size'][0]

    if gradir == 1:
     flux_dir = [0,1,0]
     length = data['size'][1]

    if gradir == 2:
     flux_dir = [0,0,1]
     length = data['size'][2]

    delta = 1e-2
    nsides = len(data['sides'])
    side_value = np.zeros(nsides)

    tmp = list(data['periodic_sides']) + list(data['inactive_sides'])

    DeltaT = 1
    for kl,ll in enumerate(tmp) :
     Ee1,e2 = side_elem_map[ll]
     normal = face_normals[ll]  
     tmp = np.dot(normal,flux_dir)


     if tmp < - delta :
        side_value[ll] = -DeltaT
     if tmp > delta :
        side_value[ll] = +DeltaT
     

    side_periodic_value = np.zeros((nsides,2))
    periodic_values = {}
    periodic_side_values = {}

    n_el = len(data['elems'])
    B = sp.dok_matrix((n_el,n_el),dtype=np.float64)

    B_with_area_old = sp.dok_matrix((n_el,n_el),dtype=np.float64)
    B_area = np.zeros(n_el,dtype=np.float64)
   
    counter = 0
    pp = ()
    ip = []; jp = []; dp = []; pv = [] 
    if len(data['periodic_sides']) > 0:
     for side in data['pairs']:

      area = data['areas'][side[0]]
      
      side_periodic_value[side[0]][0] = side_value[side[0]]
      side_periodic_value[side[0]][1] = side_value[side[1]]

      i,j = side_elem_map[side[0]]

      periodic_values.update({i:{j:[side_value[side[0]]]}})
      periodic_values.update({j:{i:[side_value[side[1]]]}})
    
      periodic_side_values.update({side[0]:side_value[side[0]]})

      normal = data['face_normals'][side[0]]
      
      voli = volumes[i]
      volj = volumes[j]
 
      B[i,j] = side_value[side[0]]
      B[j,i] = side_value[side[1]]
  

      if abs(side_value[side[0]]) > 0:
       pp += ((data['ij'].index([i,j]),side_value[side[0]]),)
       ip.append(i); jp.append(j); dp.append(side_value[side[0]]); pv.append(normal*area/voli)
       counter +=1

      if abs(side_value[side[1]]) > 0:
       pp += ((data['ij'].index([j,i]),side_value[side[1]]),)
       ip.append(j); jp.append(i); dp.append(side_value[side[1]]); pv.append(-normal*area/volj)
       counter +=1

      if np.linalg.norm(np.cross(data['face_normals'][side[0]],applied_grad)) < 1e-12:
       B_with_area_old[i,j] = abs(side_value[side[0]]*area)
    

    #select flux sides-------------------------------
    flux_sides = [] #where flux goes
    total_area = 0
    for ll in data['periodic_sides']:
     e1,e2 = side_elem_map[ll]
     normal = data['face_normals'][ll]   
     tmp = np.abs(np.dot(normal,flux_dir))
     if tmp > delta : #either negative or positive
       if abs(normal[0]) == 1:
           if dim == 2:  
            area_flux = data['size'][1]
           else: 
            area_flux = data['size'][1]*data['size'][2]
           total_area += data['areas'][ll]

       elif abs(normal[1]) == 1:
           if data['dim'] == 2:  
            area_flux = data['size'][0]
           else:
            area_flux = data['size'][0]*data['size'][2]
           total_area += data['areas'][ll]

       else : #along z
           area_flux = data['size'][0]*data['size'][1]
           total_area += data['areas'][ll]


       flux_sides.append(ll)

    #-------THIS
    if argv.setdefault('contact_area','box') == 'box':
      kappa_factor = data['size'][gradir]/area_flux
    else:  
      kappa_factor = data['size'][gradir]/total_area

    data['kappa_mask']= -np.array(np.sum(B_with_area_old.todense(),axis=0))[0]*kappa_factor*1e-18
    data['periodic_side_values'] = [periodic_side_values[ll]  for ll in data['periodic_sides']]
    data['pp'] = np.array(pp)
    data['applied_gradient'] = np.array(applied_grad)
    data['flux_sides'] = np.array(flux_sides)
    data['kappa_factor'] = kappa_factor



def compute_dists(data):
  

  dists_side = np.zeros((len(data['sides']),3))
  side_elem_map = data['side_elem_map_vec']
  centroids = data['centroids']

  for ll in data['active_sides']:
   if not (ll in data['boundary_sides']):
    elem_1 = side_elem_map[ll][0]
    elem_2 = side_elem_map[ll][1]
    c1 = centroids[elem_1]
    c2 = get_next_elem_centroid(elem_1,ll,data)
    dist = c2 - c1
    dists_side[ll] = dist 

  data['dists'] = np.array(dists_side) 




def compute_interpolation_weigths(data):

  #from here: http://geomalgorithms.com/a05-_intersect-1.html

  net_sides=data['active_sides'][~np.isin(data['active_sides'],data['boundary_sides'])] #Get only the nonboundary sides

  e1 =  data['side_elem_map_vec'][net_sides,0]

  w =  np.zeros((len(data['sides']),3))

  w[net_sides] = data['centroids'][e1] - data['nodes'][data['sides'][net_sides,0]]

  tmp  = np.einsum('ui,ui->u',data['face_normals'],data['dists'])
  tmp2 = np.einsum('ui,ui->u',data['face_normals'],w)

  interp_weigths = np.zeros(len(data['sides']))
  interp_weigths[net_sides]   = 1+tmp2[net_sides]/tmp[net_sides] #this is the relative distance with the centroid of the second element


  #check----------------------------------
  #ll = data['active_sides'][0]
  #(e1,_) = data['side_elem_map_vec'][ll]
  #P2 = data['nodes'][data['sides'][ll][0]]
  #P3 = data['nodes'][data['sides'][ll][1]]
  #P0 = data['centroids'][e1]
  #P1 = get_next_elem_centroid(e1,ll,data)
  #plot([P0[0],P1[0]],[P0[1],P1[1]],'r')
  #plot([P2[0],P3[0]],[P2[1],P3[1]],'b')
  #text(P0[0],P0[1],'P0')
  #text(P1[0],P1[1],'P1')
  #text(P2[0],P2[1],'P2')
  #text(P3[0],P3[1],'P3')
  #print(interp_weigths[ll])
  #show()

  if len(data['boundary_sides']) > 0:
   interp_weigths[data['boundary_sides']] = 1

  data['interp_weigths'] = interp_weigths


def compute_connecting_matrix(data):

   elems         = data['elems']
   elem_side_map = data['elem_side_map_vec']
   side_elem_map = data['side_elem_map_vec']
   elem_volumes  = data['volumes']
   side_areas    = data['areas']
   dim           = data['dim']
   face_normals  = data['face_normals']
   a_sides =       data['active_sides']
   b_sides =       data['boundary_sides']
   c_sides =       data['interface_sides']

   #Active sides-----
   net_sides=a_sides[~np.isin(a_sides,list(b_sides)+list(c_sides))]
   i2 = side_elem_map[net_sides].flatten()
   j2 = np.flip(side_elem_map[net_sides],1).flatten()
   common = np.einsum('u,ui->ui',side_areas[net_sides],face_normals[net_sides,:dim])
   t1 =  common/elem_volumes[side_elem_map[net_sides][:,0],None]
   t2 =  common/elem_volumes[side_elem_map[net_sides][:,1],None]
   k2 = np.zeros((2*t1.shape[0],dim))
   k2[np.arange(t1.shape[0])*2]   =  t1
   k2[np.arange(t1.shape[0])*2+1] = -t2
   k2 = k2.T     
   #----------------

   #ij - it will need to be removed
   ij2 = np.zeros((2*t1.shape[0],2))
   ij2[np.arange(t1.shape[0])*2  ] = side_elem_map[net_sides]
   ij2[np.arange(t1.shape[0])*2+1] = np.flip(side_elem_map[net_sides],1)

   #-------------------------------------

   #Interface sides----
   if len(c_sides) >0 :
    inti2 = side_elem_map[c_sides].flatten()
    intj2 = np.flip(side_elem_map[c_sides],1).flatten()
    tmp = np.flip(side_elem_map[c_sides],1).flatten()
    common = np.einsum('u,ui->ui',side_areas[c_sides],face_normals[c_sides,:dim])
    t1 =  common/elem_volumes[side_elem_map[c_sides][:,0],None]
    t2 =  common/elem_volumes[side_elem_map[c_sides][:,1],None]
    intk2 = np.zeros((2*t1.shape[0],dim))
    intk2[np.arange(t1.shape[0])*2]   =  t1
    intk2[np.arange(t1.shape[0])*2+1] = -t2
    intk2 = intk2.T     
   else:
    inti2 = []   
    intj2 = []   
    intk2 = []   
    
   #boundary sides----
   sb2 = a_sides[np.isin(a_sides,b_sides)]
   eb2 = side_elem_map[sb2][:,0]
   db2 = face_normals[sb2,:dim].T*side_areas[sb2]/elem_volumes[eb2]
   #-------------
   
   data['intk'] = intk2; 
   data['inti'] = inti2;
   data['intj'] = intj2;
   data['i']    = np.array(i2); 
   data['j']    = np.array(j2); 
   data['k']    = np.array(k2);
   data['eb']   = eb2; 
   data['sb']   = sb2; 
   data['db']   = db2; 
   data['ij']   = ij2.tolist(); 

def get_next_elem_centroid(elem,side,data):

  elem_centroids    = data['centroids']
  side_elem_map     = data['side_elem_map_vec']
  side_periodicity  = data['side_periodicity']

  centroid = elem_centroids[elem]

  if not (elem in side_elem_map[side]) : print('error, no neighbor',side,elem)
  for tmp in side_elem_map[side] :
       if not (tmp == elem) :
         elem2 = tmp  
         break

  centroid2 = elem_centroids[elem2]
  ind1 = list(side_elem_map[side]).index(elem)
  ind2 = list(side_elem_map[side]).index(elem2)
  centroid = centroid2 - side_periodicity[side][ind1] + side_periodicity[side][ind2]

  return centroid


def compute_least_square_weigths(data):

   #elems     = data['elems']
   side_elem_map = data['side_elem_map_vec']
   elem_side_map = data['elem_side_map_vec']
   elem_centroids = data['centroids']
   side_centroids = data['side_centroids']
   dists = data['dists']
   dim = data['dim']
   n_elems = len(data['elems'])


   diff_dist = np.zeros((n_elems,max(data['side_per_elem']),dim))

   for ll in data['active_sides'] :
    elems = side_elem_map[ll]
    dist = dists[ll]
    kc1 = elems[0]
    ind1 = list(elem_side_map[kc1]).index(ll)
    if not ll in data['boundary_sides'] :
     kc2 = elems[1]
     ind2 = list(elem_side_map[kc2]).index(ll)
     diff_dist[kc1,ind1] =  dist[:dim]
     diff_dist[kc2,ind2] = -dist[:dim]
    else :
     dist = side_centroids[ll] - elem_centroids[kc1]
     diff_dist[kc1,ind1] = dist[:dim]

   #We solve the pinv for a stack of matrices. We do so for each element group

   index3 = np.where(data['side_per_elem'] == 3)[0]
   G3 = np.linalg.pinv(diff_dist[index3,:3])
   

   index4 = np.where(data['side_per_elem'] == 4)[0]
   G4 = np.linalg.pinv(diff_dist[index4,:4])

   G = np.zeros((n_elems,dim,max(data['side_per_elem'])))  

   G[index3,:,:3] = G3
   G[index4,:,:4] = G4

   data.update({'weigths':G})


def compute_boundary_connection(data):


     elem_centroids = data['centroids']
     side_centroids = data['side_centroids']
     side_elem_map = data['side_elem_map_vec']
     face_normals = data['face_normals']
     

     bconn = []
     for s in data['boundary_sides']:
        ce = elem_centroids[side_elem_map[s][0]]
        cs = side_centroids[s]
        d = cs-ce
        normal = face_normals[s]
        d1 = normal * np.dot(normal,d)
        bconn.append(d1)

     data.update({'bconn':bconn})


def compute_face_normals(data):


   sides = data['sides']
   nodes = data['nodes']
   dim = data['dim']
   side_elem_map = np.array(data['side_elem_map_vec'])
   elem_centroids = np.array(data['centroids'])
   side_centroids = np.array(data['side_centroids'])

   a= time.time()
   v1 =  nodes[sides[:,1]]-nodes[sides[:,0]]

   if dim == 3:
    v2 =  nodes[sides[:,2]]-nodes[sides[:,1]]
   else :
    v2 = np.array([0,0,1]).T
   v = np.cross(v1,v2)

   normal = v.T/np.linalg.norm(v,axis=1)
   
   c = side_centroids - elem_centroids[side_elem_map[:,0]]   
   index = np.where(np.einsum('iu,ui->u',normal,c) < 0)[0]
   normal[:,index] = -normal[:,index]
   normal = normal.T
 
   #---------------------------------------------------------
   data.update({'face_normals':normal})


def import_mesh(**argv):

   data = {}

   elem_region_map = {}
   region_elem_map = {}

   with open('mesh.msh', 'r') as f: lines = f.readlines()

   lines = [l.split()  for l in lines]
   nb = int(lines[4][0])
   current_line = 5
   blabels = {int(lines[current_line+i][1]) : lines[current_line+i][2].replace('"',r'') for i in range(nb)}

   current_line += nb+2
   n_nodes = int(lines[current_line][0])
   current_line += 1
   nodes = np.array([lines[current_line + n][1:4] for n in range(n_nodes)],float)
   current_line += n_nodes+1

   size = [ np.max(nodes[:,i]) - np.min(nodes[:,i])   for i in range(3)]
   
   dim = 2 if size[2] == 0 else 3
   current_line += 1
   n_elem_tot = int(lines[current_line][0])
   current_line += 1

   #type of elements
   bulk_type = {2:[2,3],3:[4]}
   face_type = {2:[1],3:[2]}
   #---------------------------

   bulk_tags = [n for n in range(n_elem_tot) if int(lines[current_line + n][1]) in bulk_type[dim]] 
   face_tags = [n for n in range(n_elem_tot) if int(lines[current_line + n][1]) in face_type[dim]]
   elems = [list(np.array(lines[current_line + n][5:],dtype=int)-1) for n in bulk_tags]
   side_per_elem = np.array([len(e) for e in elems])
   n_elems = len(elems)
   boundary_sides = np.array([ sorted(np.array(lines[current_line + n][5:],dtype=int)) for n in face_tags] ) -1

   #generate sides and maps---
   node_side_map = { i:[] for i in range(len(nodes))}
   elem_side_map = { i:[] for i in range(len(elems))}
   sides = []
   tmp_indices = []
   for k,elem in enumerate(elems):
       tmp = list(elem)
       for i in range(len(elem)):
         tmp.append(tmp.pop(0))
         trial = sorted(tmp[:dim])
         sides.append(trial)
         tmp_indices.append(k)
   sides,inverse = np.unique(sides,axis=0,return_inverse = True)
   for k,s in enumerate(inverse): elem_side_map[tmp_indices[k]].append(s)

   for s,side in enumerate(sides): 
        for t in side:
            node_side_map[t].append(s)
   side_elem_map = { i:[] for i in range(len(sides))}
   for key, value in elem_side_map.items():
     for v in value:   
      side_elem_map[v].append(key)    
   #----------------------------------------------------      
 
   #Build relevant data
   side_areas = compute_side_areas(nodes,sides,dim) 
 
   #Compute volumes
   elem_volumes = compute_elem_volumes(nodes,elems,dim)

   data.update({'elems':elems,'sides':sides,'nodes':nodes,'volumes':elem_volumes})

   side_centroids = np.array([np.mean(nodes[i],axis=0) for i in data['sides']] )
   elem_centroids = np.array([np.mean(nodes[i],axis=0) for i in data['elems']])

   data.update({'side_centroids':side_centroids})
   data.update({'centroids':elem_centroids})
   
   data.update({'side_per_elem':np.array(side_per_elem)})


   #match the boundary sides with the global side.
   physical_boundary = {}
   for n,bs in enumerate(boundary_sides): #match with the boundary side
      side = node_side_map[bs[0]]
      for s in side: 
        if np.allclose(np.array(sides[s]),np.array(bs),rtol=1e-4,atol=1e-4):
            physical_boundary.setdefault(blabels[int(lines[current_line + n][3])],[]).append(s) 
            break

   side_list = {}
   #Apply Periodic Boundary Conditions
   side_list.update({'active':list(range(len(sides)))})
   side_periodicity = np.zeros((len(sides),2,3))
   group_1 = []
   group_2 = []
   #self.pairs = [] #global (all periodic pairs)

   side_list.setdefault('Boundary',[])
   side_list.setdefault('Interface',[])
   periodic_nodes = {}

   if argv.setdefault('delete_gmsh_files',False):
    os.remove(os.getcwd() + '/mesh.msh')
    os.remove(os.getcwd() + '/mesh.geo')

   pairs = []
   for label in list(physical_boundary.keys()):

    if str(label.split('_')[0]) == 'Periodic':
     if not int(label.split('_')[1])%2==0:
      contact_1 = label
      contact_2 = 'Periodic_' + str(int(label.split('_')[1])+1)
      group_1 = physical_boundary[contact_1]
    
      group_2 = physical_boundary[contact_2]
      for s in group_1:
        c = side_centroids[s]
      for s in group_2:
        c = side_centroids[s]

      #----create pairs
      new_pairs = []
      #compute tangential unity vector
      tmp = nodes[sides[group_2[0]][0]] - nodes[sides[group_2[0]][1]]
      t = tmp/np.linalg.norm(tmp)
      n = len(group_1)
      for s1 in group_1:
       d_min = 1e6
       for s in group_2:
        c1 = side_centroids[s1]
        c2 = side_centroids[s]
        d = np.linalg.norm(c2-c1)
        if d < d_min:
         d_min = d
         pp = c1-c2
         s2 = s
       new_pairs.append([s1,s2])
       side_periodicity[s1][1] = pp
       side_periodicity[s2][1] = -pp
      pairs +=new_pairs
      #----------------------------------
      #Amend map
      for s in new_pairs:
       s1 = s[0]
       s2 = s[1]

       #Change side in elem 2--------------------
       elem2 = side_elem_map[s2][0]
       index = elem_side_map[elem2].index(s2)
       elem_side_map[elem2][index] = s1
    
       side_elem_map[s1].append(elem2)
       side_elem_map[s2].append(side_elem_map[s1][0])
       side_list['active'].remove(s2)
       #-----------------------------------------
        
      #Polish sides
      [side_list.setdefault('Periodic',[]).append(i) for i in physical_boundary[contact_1]]
      [side_list.setdefault('Inactive',[]).append(i) for i in physical_boundary[contact_2]]

   if 'Boundary' in physical_boundary.keys(): side_list.update({'Boundary':physical_boundary['Boundary']})

   for side in side_list['Boundary'] :# + self.side_list['Hot'] + self.side_list['Cold']:
    side_elem_map[side].append(side_elem_map[side][0])

   #Put boundaries at the end
   n_non_boundary_side_per_elem = []
   for key,value in elem_side_map.items():
       tmp_1 = []
       tmp_2 = []
       for s in value:
         if s in side_list['Boundary']:
           tmp_1.append(s)
         else:  
           tmp_2.append(s)
       n_non_boundary_side_per_elem.append(len(tmp_2)+len(tmp_1))
       elem_side_map[key] = tmp_2 + tmp_1


   #Fill with -1 the missing connectivity
   ms = max(side_per_elem)
   for n,e in enumerate(elems):
        if len(e) < ms:
            e.append(-1)
            elem_side_map[n].append(-1)
   #----------------------------------------


   data.update({'active_sides':np.array(side_list['active'])})
   data.update({'n_non_boundary_side_per_elem':np.array(n_non_boundary_side_per_elem)})
   data.update({'boundary_sides':np.array(side_list['Boundary'])})
   data.update({'periodic_sides':np.array(side_list['Periodic'])})
   data.update({'side_periodicity':side_periodicity})
   data.update({'areas':np.array(side_areas)})
   data.update({'elem_mat_map':np.zeros(len(elems))})

   data.update({'dim':dim})
   data.update({'dmin':argv['dmin']})
   data.update({'pairs':pairs})
   #aa = np.array([side_elem_map[ll]  for ll in range(len(sides))]) 
   data.update({'elem_side_map_vec': np.array([elem_side_map[ll]  for ll in range(len(elems))]) })
   data.update({'side_elem_map_vec': np.array([side_elem_map[ll]  for ll in range(len(sides))]) })
   data['interface_sides']= []
   direction = argv.setdefault('direction','x')
   if direction == 'x':
    data['direction'] = 0
   elif  direction == 'y': 
    data['direction'] = 1
   else: 
    data['direction'] = 2
   
   data['size'] = size
   data['inactive_sides'] = side_list['Inactive']

   return data
   #------------------


def compute_data(data,**argv):

   compute_face_normals(data)

   compute_boundary_connection(data)
  
   compute_dists(data)
   
   compute_least_square_weigths(data)
   
   compute_connecting_matrix(data)

   compute_interpolation_weigths(data)
   
   compute_boundary_condition_data(data,**argv)

   #Some adjustement--


   data['meta'] = np.asarray([len(data['elems']),data['kappa_factor'],data['dim'],len(data['nodes']),len(data['active_sides']),data['direction'],data['dmin']],np.float64)
   
   del data['direction']
   del data['dim']
   del data['dmin']
   del data['kappa_factor']




def Geometry(**argv):

 if comm.rank == 0 :

   if not argv.setdefault('create_mesh',False):  
    Mesher(argv) 
   data = import_mesh(**argv)

   compute_data(data,**argv)

   if argv.setdefault('save',True):
     save_data(argv.setdefault('output_filename','geometry'),data)   

   return data



def compute_elem_volumes(nodes,elems,dim):


  n_elems = len(elems)
  elem_volumes = np.zeros(len(elems))
  if dim == 3: #Assuming Tetraedron
   M = np.ones((4,4))
   for i,e in enumerate(elems):
    M[:,:3] = nodes[e,:]
    elem_volumes[i] = 1.0/6.0 * abs(np.linalg.det(M))

  if dim == 2: 
   M = np.ones((3,3))
   for i,e in enumerate(elems):
    if len(e) == 3:     
     M[:,:2] = nodes[e,0:2]
     elem_volumes[i] = abs(0.5*np.linalg.det(M))
    else:

     points = nodes[e]
     x = points[:,0]
     y = points[:,1]
     elem_volumes[i] = 0.5*np.abs(np.dot(x,np.roll(y,1))-np.dot(y,np.roll(x,1)))

  return elem_volumes   

comm = MPI.COMM_WORLD


def compute_side_areas(nodes,sides,dim):

  if dim == 2:
   side_areas = [  np.linalg.norm(nodes[s[1]] - nodes[s[0]]) for s in sides ]
  else:     
   p =nodes[sides]
   side_areas = np.linalg.norm(np.cross(p[:,0]-p[:,1],p[:,0]-p[:,2]),keepdims=True,axis=1).T[0]/2

  return side_areas 





        




import numpy as np
import math
from .utils import *


def get_shape(argv):

    
    #Vectorize shape if needed
    shape = argv.setdefault('shape','square') 
    if type(shape) == list:
     assert(len(shape)==len(argv['base']))
     shapes_str = shape
    else:
     shapes_str = len(argv['base'])*[shape]   

    #Vectorize custom shapes if needed
    n_custom = 0
    for shape in shapes_str:
        if shape == 'custom':
           n_custom +=1

    if n_custom > 0:
      if type(argv['shape_function']) == list:
       assert(len(argv['shape_function'])==len(argv['base']))
      else:
       argv['shape_function'] = len(argv['base'])*[argv['shape_function']]   
    #-----------------------
           

    #compute area weigths
    argv.setdefault('area_ratio',np.ones(len(argv['base'])))

    if argv.setdefault('relative',True):
     areas =  [argv['porosity']* i /sum(argv['area_ratio'])  for i in argv['area_ratio']]
    else: 
     areas = argv['area_ratio']

    #-------------------
    shapes = []
    n_custom = 0
    for n,shape in enumerate(shapes_str):

     area= areas[n]

     if shape == 'square':
       shapes.append(np.array(make_polygon(4,area)))

     if shape == 'triangle':
       shapes.append(np.array(make_polygon(3,area)))

     elif shape == 'circle':
       shapes.append(np.array(make_polygon(24,area)))

     elif shape == 'custom':

      options = {key:value[n_custom] if isinstance(value,list) else value for key,value in argv.setdefault('shape_options',{}).items()}

      options.update({'area':area})

      shapes.append(argv['shape_function'][n_custom](**options))
      n_custom +=1


    return shapes  






def get_smoothed_square(**argv):

     
     smooth = argv['smooth']
     area = argv['area']
     Na = argv['Na']
     L = np.sqrt(area+smooth*smooth*(4-np.pi))

     p = []
     dphi = math.pi/Na/2;
     for ka in range(Na+1):
      ph =  dphi * ka
      px  = L/2 - smooth + smooth * math.cos(ph)
      py  = L/2 - smooth + smooth * math.sin(ph)
      p.append([px,py])

     dphi = math.pi/Na/2;
     for ka in range(Na+1):
      ph = np.pi/2 +  dphi * ka
      px  = - L/2 + smooth + smooth * math.cos(ph)
      py  = + L/2 - smooth + smooth * math.sin(ph)
      p.append([px,py])

     dphi = math.pi/Na/2;
     for ka in range(Na+1):
      ph = np.pi +  dphi * ka
      px  = - L/2 + smooth + smooth * math.cos(ph)
      py  = - L/2 + smooth + smooth * math.sin(ph)
      p.append([px,py])

     dphi = math.pi/Na/2;
     for ka in range(Na+1):
      ph = 1.5*np.pi +  dphi * ka
      px  = + L/2 - smooth + smooth * math.cos(ph)
      py  = - L/2 + smooth + smooth * math.sin(ph)
      p.append([px,py])

     return p










import sys
import numpy as np
from .utils import *


def gray2D(**argv):

 #Import data----
 kappa_bulk = argv['kappa']
 kappa = argv['kappa']*np.eye(2)
 mfp_bulk = argv['mfp']
 n_phi = int(argv.setdefault('n_phi',48))
 Dphi = 2*np.pi/n_phi
 phi = np.linspace(Dphi/2.0,2.0*np.pi-Dphi/2.0,n_phi,endpoint=True)
 polar = np.array([np.sin(phi),np.cos(phi)]).T
 fphi= np.sinc(Dphi/2.0/np.pi)
 polar_ave = polar*fphi
 #------------------

 kappa_directional = np.zeros((1,n_phi,2))
 temp_coeff = np.zeros((1,n_phi))
 temp_coeff[0,:] = 1/n_phi
 suppression = np.zeros((1,n_phi,1)) 
 for p in range(n_phi): 
    kappa_directional[0,p] += kappa_bulk/mfp_bulk * polar_ave[p]/n_phi*2

 tc = temp_coeff/np.sum(temp_coeff)


 #Final----
 return {'tc':tc,\
         'sigma':kappa_directional,\
         'kappa':kappa,\
         'mfp_average':np.array([mfp_bulk]),\
         'VMFP':polar_ave,\
         'mfp_sampled':np.array([mfp_bulk]),\
         'model':np.array([1]),\
         'suppression':np.zeros(1),\
         'phi': phi  ,\
         'directions': polar  ,\
         'kappam':np.array([kappa_bulk])}





from openbte.utils import *



def read_data(**argv):

    data = {}
    if len(argv) == 0:
     for name in sys.argv[1:]:
       tmp = load_data(name[:-4])
       data[name[:-4]] = tmp
    else:   
     for name,value in argv.items():
       tmp = load_data(name)
       data[name] = value

    return data   




def main():

 
 data = read_data()  


 save_data('bundle',data)













import sys
import numpy as np
from .utils import *


def gray2DSym(**argv):

 #Import data----
 #kappa_bulk = argv['kappa']
 #kappa = argv['kappa']*np.eye(2)
 #mfp_bulk = argv['mfp']
 #n_phi = int(argv.setdefault('n_phi',48)) 
 #n_mfp = int(argv.setdefault('n_mfp',50)) 
 #n_theta = int(argv.setdefault('n_theta',48))
 #Dphi = 2*np.pi/n_phi
 #phi = np.linspace(Dphi/2.0,2.0*np.pi-Dphi/2.0,n_phi,endpoint=True)
 #polar = np.array([np.sin(phi),np.cos(phi)]).T
 #Dtheta = np.pi/n_theta
 #theta = np.linspace(Dtheta/2.0,np.pi - Dtheta/2.0,n_theta)
 #dtheta = 2.0*np.sin(Dtheta/2.0)*np.sin(theta)   
 #domega = np.outer(dtheta,Dphi * np.ones(n_phi))
 #ftheta        = (1-np.cos(2*theta)*np.sinc(Dtheta/np.pi))/(np.sinc(Dtheta/2/np.pi)*(1-np.cos(2*theta)))
 #fphi          = np.sinc(Dphi/2.0/np.pi)
 #polar_ave     = polar*fphi
 #azimuthal_ave = np.array([ftheta*np.sin(theta),ftheta*np.sin(theta),np.cos(theta)]).T  
 #direction_ave = np.einsum('tj,pj->tpj',azimuthal_ave,polar_ave)
 #direction_int = np.einsum('ijl,ij->ijl',direction_ave,domega)
 #------------------------------------------------------

 kappa_bulk = argv['kappa']
 kappa = argv['kappa']*np.eye(3)
 mfp_bulk = argv['mfp']
 #kappa_bulk = np.eye(3)
 n_phi = int(argv.setdefault('n_phi',48)) 
 n_mfp = int(argv.setdefault('n_mfp',50)) #this is K
 n_theta = int(argv.setdefault('n_theta',48))
 nm = n_phi * n_mfp
 #Create sampled MFPs
 min_mfp = 1e-10 #m
 max_mfp = 1e-3  #m
 mfp = np.logspace(np.log10(min_mfp),np.log10(max_mfp),n_mfp)
 n_mfp = len(mfp)
 #Polar Angle---------
 Dphi = 2*np.pi/n_phi
 phi = np.linspace(0,2.0*np.pi,n_phi,endpoint=False)
 #Azimuthal Angle------------------------------
 Dtheta = np.pi/n_theta
 theta = np.linspace(Dtheta/2.0,np.pi - Dtheta/2.0,n_theta)
 dtheta = 2.0*np.sin(Dtheta/2.0)*np.sin(theta)   
 domega = np.outer(dtheta,Dphi * np.ones(n_phi))

 #Compute directions---
 polar = np.array([np.sin(phi),np.cos(phi),np.ones(n_phi)]).T
 azimuthal = np.array([np.sin(theta),np.sin(theta),np.cos(theta)]).T
 direction = np.einsum('tj,pj->tpj',azimuthal,polar)
 
 #Compute average---
 ftheta        = (1-np.cos(2*theta)*np.sinc(Dtheta/np.pi))/(np.sinc(Dtheta/2/np.pi)*(1-np.cos(2*theta)))
 fphi          = np.sinc(Dphi/2.0/np.pi)
 polar_ave     = polar*fphi
 azimuthal_ave = np.array([ftheta*np.sin(theta),ftheta*np.sin(theta),np.cos(theta)]).T  
 direction_ave = np.einsum('tj,pj->tpj',azimuthal_ave,polar_ave)
 direction_int = np.einsum('ijl,ij->ijl',direction_ave,domega)

 #-----------------

 #Gray 2D Sym--------
 n_tot = n_phi*n_theta
 kappa_directional = np.zeros((1,n_tot,3))
 temp_coeff = np.zeros((1,n_tot))
 dirr = np.zeros((n_tot,3))
 for p in range(n_phi): 
  for t in range(n_theta):
    index = n_phi * t + p  
    kappa_directional[0,index] += kappa_bulk/mfp_bulk * direction_ave[t,p]/(4.0*np.pi)*3
    temp_coeff[0,index] = domega[t,p]
 #-----------------

 #Gray 2D--------
 #n_tot = n_phi
 #kappa_directional = np.zeros((1,n_tot,3))
 #temp_coeff = np.zeros((1,n_tot))
 #dirr = np.zeros((n_tot,3))
 #for p in range(n_tot): 
 #   index = p 
 #   kappa_directional[0,p] += kappa_bulk/mfp_bulk * polar_ave[p]*Dphi/(2*np.pi)*2
 #   dirr[index] = polar_ave[p]
 #   temp_coeff[0,index] = Dphi
 #-----------------




 tc = temp_coeff/np.sum(temp_coeff)

 #Final----
 return {'tc':tc,\
         'sigma':kappa_directional[:,:,:2],\
         'kappa':kappa,\
         'mfp_average':np.array([mfp_bulk]),\
         'VMFP':dirr[:,:2],\
         'mfp_sampled':np.array([mfp_bulk]),\
         'model':np.array([1]),\
         'suppression':np.zeros(1),\
         'phi': phi  ,\
         'directions': polar  ,\
         'kappam':np.array([kappa_bulk])}





import plotly.offline as py
import plotly.graph_objs as go
import numpy as np
import plotly.io as pio
import dash
import dash_core_components as dcc
import dash_html_components as html
from threading import Timer
import dash_bootstrap_components as dbc 
import flask
from dash.dependencies import Input, Output
import gzip
import pickle
import plotly.express as px
import pandas as pd
import json
import os
from mpi4py import MPI

comm = MPI.COMM_WORLD

def load_data(fh):

 if os.path.isfile(fh + '.npz'):
      with gzip.open(fh + '.npz', 'rb') as f:

          return pickle.load(f)


def get_layout(size = []):


 sl = True

 dim = 2   if len(size) == 2 else 3
 x = 1;y=size[1]/size[0]
 z= 1 if dim == 2 else size[2]/size[0]

 axis = dict(
          backgroundcolor="rgb(230, 230,230)",
          gridcolor="rgb(255, 255, 255)",
          zerolinecolor="rgb(255, 255, 255)",
          showbackground=False
         )

 xaxis = axis.copy();xaxis['title']=' x [nm]'
 yaxis = axis.copy();yaxis['title']=' y [nm]'
 zaxis = axis.copy();zaxis['title']=' z [nm]'

 font=dict(
        family="Courier New, monospace",
        size=12,
        #color="#7f7f7f"
        color="gray"
       )

 # Update 3D scene options
 scene = dict(
    aspectratio=dict(x= x, \
                     y= y, \
                     z= z),
    xaxis= xaxis,
    yaxis= yaxis,
    zaxis= zaxis)


 #camera
 camera = dict(
    up=dict(x=0, y=1, z=0),
    center=dict(x=0, y=0, z=0),
    eye=dict(x=0, y=0, z=1.5)
   )


 layout = dict(font=font,\
               #title=title,\
               autosize=True,\
               scene=scene,\
               uirevision='static',\
               scene_camera = camera,\
               xaxis_showgrid=False,\
               yaxis_showgrid=False,\
               margin =dict(t=50, b=20, l=20, r=20),\
               paper_bgcolor='rgba(0,0,0,0)',\
               plot_bgcolor='rgba(0,0,0,0)')

 return layout


def App(**argv):

 if comm.rank == 0:
  #load data--
  data_tot     = argv['bundle'] if 'bundle' in argv.keys() else load_data('bundle')



  d_sample = html.Div([
    dcc.Dropdown(
        id='samples',
        value=0,
        style = {},
        clearable=False,
        searchable=False
    ),
  ])


  d_variable = html.Div([
    dcc.Dropdown(
        id='variables',
        value=0,
        style= {},
        clearable=False,
        searchable=False
    ),
  ])


  external_stylesheets=[dbc.themes.CYBORG]
  server = flask.Flask(__name__) # define flask app.server
  app = dash.Dash(__name__, external_stylesheets=external_stylesheets,server=server)

  #Init figure---
  fig = go.Figure()
  fig.update_xaxes(showline=False,gridcolor='rgba(0,0,0,0)',zerolinecolor='rgba(0,0,0,0)')
  fig.update_yaxes(showline=False,gridcolor='rgba(0,0,0,0)',zerolinecolor='rgba(0,0,0,0)')

  fig.update_layout(
    plot_bgcolor='rgba(0,0,0,0)',
    paper_bgcolor='rgba(0,0,0,0)',
    font_color='rgba(0,0,0,0)',
  )
  #-------------

  fig.update_xaxes(showline=False)

  fig_dcc = dcc.Graph(id='master',figure=fig,style={'left':'15%','width':'70%','height':'70%','bottom':'10%','position':'absolute'})




  dd_sample = html.Div( children = [d_sample],style = {'left':'1%','height':'10%','width':'25%','bottom':'88%','position':'absolute'})
  dd_variable = html.Div( children = [d_variable],style = {'left':'28%','height':'10%','width':'25%','bottom':'88%','position':'absolute'})

  hidden = html.Div(id='intermediate-value', style={'display': 'none'})

  url = dcc.Location(id='url', refresh=True)


  img = html.Div(html.Img(src=app.get_asset_url('logo.png'), style={'width':'18%','left':'80%','bottom':'88%','position':'absolute'}))

  #div1 = html.Div(style = {'left':'25%','height':'70%','width':'50%','bottom':'15%',"border":"2px white solid",'position':'absolute'},children=[fig_dcc,dd_sample,dd_variable,hidden,url]);

  #app.layout = html.Div(children = [div1,img],style = {'height': '100vh','width': '100%',"border":"2px white solid"})
  #app.layout = html.Div(children = [fig_dcc,dd_sample,dd_variable,hidden,url  ,img],style = {'height': '100vh','width': '100%',"border":"2px white solid"})
  layout = html.Div(children = [fig_dcc,dd_sample,dd_variable,hidden,url  ,img],style = {'height': '100vh','width': '100%','position':'absolute'})

  #---------------------------------------------------------------------


  @app.callback(
    Output('master','figure'),
    Input('samples','value'), Input('variables','value'),Input('intermediate-value','children'),prevent_initial_call=True
  )
  def plot(sample,variable,data_tot):


       key_sample = list(data_tot.keys())[int(sample)]
       data = data_tot[key_sample]
       name = list(data['variables'].keys())[int(variable)]
       variable  = data['variables'][name]
       elems     = np.array(data['elems'])
       nodes     = np.array(data['nodes'])
       node_data      = variable['data']


       bb = str(round(data['bte'],2))+' W/m/K' if 'bte' in data.keys() else '--'
       text  = 'Bulk: ' + str(round(data['bulk'],2)) +' W/m/K<br>Fourier: '\
                    +       str(round(data['fourier'],2)) + ' W/m/K<br>BTE:' \
                    + bb +'<br>' + \
                    'Gradient applied along ' +  data['direction']



       z = np.zeros(len(nodes))  if np.shape(nodes)[1] == 2 else nodes[:,2]

       plotly_data = go.Mesh3d(x=nodes[:,0],
                         y=nodes[:,1],
                         z=z,
                         colorscale='Viridis',
                         colorbar_title = '[' + variable['units'] + ']' if len(variable['units']) > 0 else "",
                         intensity = node_data,
                         colorbar=dict(len=0.5),
                         intensitymode='vertex',
                         i=elems[:,0],
                         j=elems[:,1],
                         k=elems[:,2],
                         name=name,
                         showscale = True,
                         visible=True)


       fig = go.Figure(data=plotly_data,layout=get_layout(data['size']))

       fig.add_annotation(
            x=1,
            y=0.05,
            xref='paper',
            showarrow=False,
            yref='paper',
            text=text,align='left')


       return  fig



  @app.callback(
    [Output('intermediate-value', 'children'),Output('samples','options'),Output('variables','options')],
    [Input('url','search')],prevent_initial_call=True
  )
  def update_output_div(pathname):
  
    input_value = pathname[1:]

    variables = data_tot[list(data_tot.keys())[0]]['variables'].keys()

    options_samples=[{'label': s, 'value': str(n)} for n,s in enumerate(data_tot)]
    options_variables=[{'label': s, 'value': str(n)} for n,s in enumerate(variables)]
 


    return data_tot,options_samples,options_variables


  app.layout = layout
  if argv.setdefault('show',False):
   app.run_server(host='localhost',port=8000)

  return app






from __future__ import absolute_import
import numpy as np
from scipy.sparse.linalg import splu
from termcolor import colored, cprint 
from .utils import *
from .fourier import *
from mpi4py import MPI
import scipy.sparse as sp
import time
import sys
import scipy
from cachetools import cached,LRUCache
from cachetools.keys import hashkey

comm = MPI.COMM_WORLD
      
cache_compute_lu = LRUCache(maxsize=10000)

def clear_BTE_cache():

  cache_compute_lu.clear()

      
def get_m(Master,data):
    
    Master.data = data
    return Master

def get_boundary(RHS,eb,n_elems):

    if len(eb)  > 0:
     RHS_tmp = np.zeros(n_elems) 
     np.add.at(RHS_tmp,eb,RHS)
    else: 
      return np.zeros(n_elems)  

    return RHS_tmp


def print_multiscale(n_serial,n_parallel,MM):

        total = n_serial*n_parallel

        print(flush=True)
        print('                  Multiscale Diagnostics        ''',flush=True)
        print(colored(' -----------------------------------------------------------','green'),flush=True)
        diff = int(MM[0])/total
        bal = int(MM[1])/total
        print(colored(' BTE:              ','green') + str(round((1-diff-bal)*100,2)) + ' %',flush=True)
        print(colored(' FOURIER:          ','green') + str(round(diff*100,2)) + ' %',flush=True)
        print(colored(' BALLISTIC:        ','green') + str(round(bal*100,2)) + ' %',flush=True)
        print(colored(' -----------------------------------------------------------','green'),flush=True)


@cached(cache=cache_compute_lu, key=lambda data,indices:hashkey(indices))
def compute_lu(A,indices):

 return sp.linalg.splu(A)




def solve_rta(argv):

    mesh    = argv['geometry']
    mat     = argv['material']
    fourier = argv['fourier']
    n_elems = len(mesh['elems'])
    mfp = mat['mfp_sampled']*1e9
   
    sigma = mat['sigma']*1e9
    dim = int(mesh['meta'][2])
    F = mat['VMFP']
    #Get discretization----
    n_serial,n_parallel = sigma.shape[:2]


    block =  n_parallel//comm.size
    rr = range(block*comm.rank,n_parallel) if comm.rank == comm.size-1 else range(block*comm.rank,block*(comm.rank+1))

    if comm.rank == 0 and argv['verbose']:
       print(flush=True)
       print('      Iter    Thermal Conductivity [W/m/K]      Residual ''',flush=True)
       print(colored(' -----------------------------------------------------------','green'),flush=True)

    if len(mesh['db']) > 0:
      if comm.rank == 0: 
       Gb   = np.einsum('mqj,jn->mqn',mat['sigma'],mesh['db'],optimize=True)
       Gbp2 = Gb.clip(min=0);
       with np.errstate(divide='ignore', invalid='ignore'):
          tot = 1/Gb.clip(max=0).sum(axis=0).sum(axis=0); tot[np.isinf(tot)] = 0
          data = {'GG': np.einsum('mqs,s->mqs',Gbp2,tot)}
       del tot, Gbp2
      else: data = None
      if comm.size > 1:
       GG = create_shared_memory_dict(data)['GG']
      else: 
       GG = data['GG']
 
    #Bulk properties---
    G = np.einsum('qj,jn->qn',F[rr],mesh['k'],optimize=True)
    Gp = G.clip(min=0); Gm = G.clip(max=0)
    D = np.zeros((len(rr),len(mesh['elems'])))
    np.add.at(D.T,mesh['i'],Gp.T)
    DeltaT = fourier['temperature'].copy()
    #--------------------------

    #---boundary----------------------
    if len(mesh['db']) > 0: #boundary
     tmp = np.einsum('rj,jn->rn',mat['VMFP'][rr],mesh['db'],optimize=True)  
     Gbp2 = tmp.clip(min=0); Gbm2 = tmp.clip(max=0);
     np.add.at(D.T,mesh['eb'],Gbp2.T)
     TB = DeltaT[mesh['eb']]

    #Periodic---
    sss = np.asarray(mesh['pp'][:,0],dtype=int)
    P = np.zeros((len(rr),n_elems))
    np.add.at(P.T,mesh['i'][sss],-(mesh['pp'][:,1]*Gm[:,sss]).T)
    del G,Gp
    #----------------------------------------------------

    def sparse_dense_product(i,j,B,X):
     '''
     This solves B_ucc' X_uc' -> A_uc

     B_ucc' : sparse in cc' and dense in u. Data is its vectorized data COO descrition
     X      : dense matrix
     '''

     tmp = np.zeros_like(X)
     np.add.at(tmp.T,i,B.T * X.T[j])   

     return tmp

    def compute_upwind_projected_gradient(X,m,q):

         return sparse_dense_product(mesh['i'],mesh['j'],Gm[q],X) + np.einsum('c,c->c',D[q],X) - P[q]

    def compute_partial_residual(X,m,q,B):

       R = mfp[m]*compute_upwind_projected_gradient(X,m,q) + X - B

       return np.linalg.norm(R)/np.linalg.norm(X)
       

    def compute_residual(X):

       R = np.zeros_like(X)

       DeltaT = np.zeros(n_elems)
       TB = np.zeros(len(mesh['eb']))
       for m in range(n_serial):
         for q in range(n_parallel):
           DeltaT += mat['tc'][m,q]*X[q,m]
           np.add.at(TB,np.arange(mesh['eb'].shape[0]),-X[q,m,mesh['eb']]*GG[m,q])
      
       #Boundary----------
       RHS = -np.einsum('c,nc->cn',TB,Gbm2) 
       B   =  np.zeros((n_elems,n_parallel))
       np.add.at(B,mesh['eb'],RHS)
       #------------------

       for m in range(n_serial):
           R[:,m,:] = (sparse_dense_product(mesh['i'],mesh['j'],Gm,X[:,m,:]) + np.einsum('uc,uc->uc',D,X[:,m,:]) - P - B.T) + (X[:,m,:]-DeltaT)/mfp[m]

       R  = R.reshape((n_serial*n_parallel,n_elems))

       return np.linalg.norm(R,ord='fro')


    kappa_vec = [fourier['meta'][0]]
    kappa_old = kappa_vec[-1]
    error = 1
    kk = 0
    kappa_tot = np.zeros(1)
    kappaf_tot = np.zeros(1)
    MM = np.zeros(2)
    Mp = np.zeros(2)
    kappa,kappap = np.zeros((2,n_serial,n_parallel))
    kappaf,kappap_f = np.zeros((2,n_serial,n_parallel))
    kappa_0,kappap_0 = np.zeros((2,n_serial,n_parallel))
    kappab,kappap_b = np.zeros((2,n_serial,n_parallel))
    transitionp = np.zeros(n_parallel)
    transition = 1e4 * np.ones(n_parallel)
    mat['tc'] = mat['tc']/np.sum(mat['tc'])
    im = np.concatenate((mesh['i'],list(np.arange(n_elems))))
    jm = np.concatenate((mesh['j'],list(np.arange(n_elems))))
    Master = sp.csc_matrix((np.arange(len(im))+1,(im,jm)),shape=(n_elems,n_elems),dtype=np.float64)
    conversion = np.asarray(Master.data-1,np.int)
    J = np.zeros((n_elems,dim))

   
    lu = {}

    def solve(argv,A,B,indices):

     if not argv['keep_lu']:
         X = sp.linalg.spsolve(A,B)
     else: 
         
         if indices in lu:
            ll = lu[indices]
         else:   
            ll = sp.linalg.splu(A)
            lu[indices] = ll

         X = ll.solve(B)

     return X 

    gradT = compute_grad(fourier['temperature'],**argv)  

    #a = np.einsum('m,ui,ci->umc',mfp,F,gradT)
    #X_tot  = fourier['temperature'][None,None,...] - a
    #X_tot = np.tile(fourier['temperature'],(n_parallel,n_serial,1))
    #compute_residual(X_tot)


    #quit()

    while kk < argv['max_bte_iter'] and error > argv['max_bte_error']:
        a = time.time()
        if argv['multiscale']: tf,tfg = solve_fourier(mat['mfp_average'],DeltaT,argv)

        #Multiscale scheme-----------------------------
        diffusive = 0
        bal = 0
        DeltaTp = np.zeros_like(DeltaT)
        TBp = np.zeros_like(TB)
        Jp = np.zeros_like(J)
        kappa_bal,kappa_balp = np.zeros((2,n_parallel))

        RHS = -np.einsum('c,nc->nc',TB,Gbm2) if len(mesh['db']) > 0 else  np.zeros((argv['n_parallel'],0)) 

        for n,q in enumerate(rr):
           
           error_old = 1e24        
           #Get ballistic kappa
           kappap_0[:,q]   =   -np.einsum('c,m,i,ci->m',mesh['kappa_mask'],mfp,mat['VMFP'][q,0:dim],compute_grad(DeltaT,**argv))
           if argv['multiscale']:
              kappap_f[:,q] =    np.einsum('c,mc->m',mesh['kappa_mask'],tf) - np.einsum('c,m,i,mci->m',mesh['kappa_mask'],mfp,mat['VMFP'][q,0:dim],tfg) 
        
              #------------------------------------------------------
              B = DeltaT + mfp[-1]*(P[n] + get_boundary(RHS[n],mesh['eb'],n_elems))
              A = get_m(Master,np.concatenate((mfp[-1]*Gm[n],mfp[-1]*D[n]+np.ones(n_elems)))[conversion])
              X_bal = solve(argv,A,B,(q,-1))
              kappap_b[:,q] = np.dot(mesh['kappa_mask'],X_bal)
              #-----------------------------------------------------

              idx = np.argwhere(np.diff(np.sign(kappap_b[:,q] - kappap_f[:,q]))).flatten()+1
              if len(idx) == 0:
                 idx = n_serial-1
              elif len(idx) > 1:
                 idx = idx[-1] #take the last one to be on the safe side
              else:   
                 idx = idx[0]

              idx = min([idx,n_serial])

           else: idx = n_serial-1
        
           for m in range(n_serial)[idx::-1]:
            
                 #----------------------------------------
                 B = DeltaT +  mfp[m]*(P[n] + get_boundary(RHS[n],mesh['eb'],n_elems))
                 data = np.concatenate((mfp[m]*Gm[n],mfp[m]*D[n]+np.ones(n_elems)))
                 A = get_m(Master,data[conversion])
                 X = solve(argv,A,B,(q,m)) 
                 kappap[m,q] = np.dot(mesh['kappa_mask'],X)
                 #-----------------------------------------
                 if argv['multiscale']:

                  error = abs(kappap[m,q] - kappap_f[m,q])/abs(kappap[m,q])
                  if error < argv['multiscale_error_fourier']:# and m <= transition[q]:
              
                   #mm[q] = m     
                   transitionp[q] = m if argv.setdefault('transition',False) else 1e4
   
                   #Vectorize
                   kappap[:m+1,q] = kappap_f[:m+1,q]
                   diffusive += m+1
                   Xm = tf[:m+1] - np.einsum('m,i,mci->mc',mat['mfp_sampled'][:m+1],mat['VMFP'][q,:dim],tfg[:m+1])
                   DeltaTp += np.einsum('mc,m->c',Xm,mat['tc'][:m+1,q])  
                  
                   if len(mesh['eb']) > 0:
                     np.add.at(TBp,np.arange(mesh['eb'].shape[0]),-np.einsum('mc,mc->c',Xm[:,mesh['eb']],GG[:m+1,q]))
                   
                   Jp += np.einsum('mc,mj->cj',Xm,sigma[:m+1,q,0:dim])*1e-18
                   break

                 DeltaTp += X*mat['tc'][m,q]

                 if len(mesh['eb']) > 0:
                  np.add.at(TBp,np.arange(mesh['eb'].shape[0]),-X[mesh['eb']]*GG[m,q])

                 Jp += np.einsum('c,j->cj',X,sigma[m,q,0:dim])*1e-18
                
 
           for m in range(n_serial)[idx+1:]:

               #----------------------------------------------------
               B = DeltaT +  mfp[m]*(P[n] + get_boundary(RHS[n],mesh['eb'],n_elems))
               A = get_m(Master,np.concatenate((mfp[m]*Gm[n],mfp[m]*D[n]+np.ones(n_elems)))[conversion])
               X = solve(argv,A,B,(q,m))
               kappap[m,q] = np.dot(mesh['kappa_mask'],X)
               #-------------------------------------------------------

               error_bal = abs(kappap[m,q] - kappap_b[m,q])/abs(kappap[m,q])
               if error_bal < argv['multiscale_error_ballistic'] and \
                   abs(kappap[m-1,q] - kappap_b[m-1,q])/abs(kappap[m-1,q]) < argv['multiscale_error_ballistic'] and  m > int(len(mat['mfp_sampled'])/2):

                   kappap[m:,q] = kappap_b[m:,q]
                   bal += n_serial-m
                   DeltaTp += X_bal*np.sum(mat['tc'][m:,q])
                   

                   if len(mesh['eb']) > 0:
                    np.add.at(TBp,np.arange(mesh['eb'].shape[0]),-np.einsum('c,mc->c',X_bal[mesh['eb']],GG[m:,q]))

                   Jp += np.einsum('c,j->cj',X_bal,np.sum(sigma[m:,q,0:dim],axis=0))*1e-18
                   break

               DeltaTp += X*mat['tc'][m,q]
               
               if len(mesh['eb'])>0:
                np.add.at(TBp,np.arange(mesh['eb'].shape[0]),-X[mesh['eb']]*GG[m,q])

               Jp += np.einsum('c,j->cj',X,sigma[m,q,0:dim])*1e-18

        Mp[0] = diffusive
        Mp[1] = bal

        #Centering--
        #DeltaT = DeltaT - (max(DeltaT)+min(DeltaT))/2.0

        DeltaT_old = DeltaT.copy()

        comm.Barrier()

        comm.Allreduce([DeltaTp,MPI.DOUBLE],[DeltaT,MPI.DOUBLE],op=MPI.SUM)
        comm.Allreduce([Jp,MPI.DOUBLE],[J,MPI.DOUBLE],op=MPI.SUM)
        comm.Allreduce([kappap,MPI.DOUBLE],[kappa,MPI.DOUBLE],op=MPI.SUM)
        comm.Allreduce([kappap_b,MPI.DOUBLE],[kappab,MPI.DOUBLE],op=MPI.SUM)
        comm.Allreduce([kappap_0,MPI.DOUBLE],[kappa_0,MPI.DOUBLE],op=MPI.SUM)
        comm.Allreduce([kappap_f,MPI.DOUBLE],[kappaf,MPI.DOUBLE],op=MPI.SUM)
        comm.Allreduce([Mp,MPI.DOUBLE],[MM,MPI.DOUBLE],op=MPI.SUM)
        comm.Allreduce([transitionp,MPI.DOUBLE],[transition,MPI.DOUBLE],op=MPI.SUM)
        comm.Allreduce([TBp,MPI.DOUBLE],[TB,MPI.DOUBLE],op=MPI.SUM)
        kappa_totp = np.array([np.einsum('mq,mq->',sigma[:,rr,int(mesh['meta'][-1])],kappa[:,rr])])
        kappaf_totp = np.array([np.einsum('mq,mq->',sigma[:,rr,int(mesh['meta'][-1])],kappaf[:,rr])])
        comm.Allreduce([kappa_totp,MPI.DOUBLE],[kappa_tot,MPI.DOUBLE],op=MPI.SUM)
        comm.Allreduce([kappaf_totp,MPI.DOUBLE],[kappaf_tot,MPI.DOUBLE],op=MPI.SUM)


        if argv['multiscale'] and comm.rank == 0: print_multiscale(n_serial,n_parallel,MM) 

        kk +=1
        error = abs(kappa_old-kappa_tot[0])/abs(kappa_tot[0])
        kappa_old = kappa_tot[0]
        kappa_vec.append(kappa_tot[0])
        if argv['verbose'] and comm.rank == 0:   
         print('{0:8d} {1:24.4E} {2:22.4E}'.format(kk,kappa_vec[-1],error),flush=True)
 

    if argv['verbose'] and comm.rank == 0:
      print(colored(' -----------------------------------------------------------','green'),flush=True)



    kappaf -=np.dot(mesh['kappa_mask'],DeltaT_old)
    kappa -=np.dot(mesh['kappa_mask'],DeltaT_old)
    bte =  {'kappa':kappa_vec,'temperature':DeltaT,'flux':J,'kappa_mode':kappa,'kappa_mode_f':kappaf,\
            'kappa_mode_b':kappab-np.dot(mesh['kappa_mask'],DeltaT_old),'kappa_0':kappa_0}
    argv['bte'] = bte



          
     











import sys
import numpy as np
import scipy
from mpi4py import MPI
import time
import os 
from openbte.utils import *


comm = MPI.COMM_WORLD

def get_linear_indexes(mfp,value):


   n = len(mfp)

   if value < mfp[0]:
      aj = (value-mfp[0])/(mfp[1]-mfp[0]) 
      return 0,1-aj,1,aj  
   elif value > mfp[-1]:
      aj = (value-mfp[n-2])/(mfp[n-1]-mfp[n-2]) 
      return n-2,1-aj,n-1,aj  
   else:
    for m in range(n-1):
      if (value <= mfp[m+1]) and (value >= mfp[m]) :
        aj = (value-mfp[m])/(mfp[m+1]-mfp[m]) 
        return m,1-aj,m+1,aj  


def mfp2D(**argv):

 #load data
 if argv.setdefault('read_from_file',True):
   data = load_data('mfp')
   Kacc = data['Kacc']
   mfp_bulk  = data['mfp']
 else:  
   mfp_bulk = argv['mfp']
   Kacc    = argv['Kacc']

 #Get the distribution
 kappa_bulk = Kacc[-1]
 kappa_bulk = np.zeros_like(Kacc)
 kappa_bulk[0] = Kacc[0]
 for n in range(len(Kacc)-1):
    kappa_bulk[n+1] = Kacc[n+1]-Kacc[n]
 #-------------------------------

 #pruning--
 I = np.where(mfp_bulk > 0)
 kappa_bulk = kappa_bulk[I]
 mfp_bulk = mfp_bulk[I]
 kappa= np.eye(3) * np.sum(kappa_bulk)

 #Get options----
 n_phi = int(argv.setdefault('n_phi',48))


 #Create sampled MFPs
 n_mfp_bulk = len(mfp_bulk)
 if argv.setdefault('interpolation',True):
  n_mfp = int(argv.setdefault('n_mfp',50))
  mfp_sampled = np.logspace(min([-2,np.log10(min(mfp_bulk)*0.99)]),np.log10(max(mfp_bulk)*1.01),n_mfp)#min MFP = 1e-2 nm 
 else:
  mfp_sampled = mfp_bulk
  n_mfp = len(mfp_bulk)

 nm = n_phi * n_mfp
 #Polar Angle---------
 Dphi = 2*np.pi/n_phi
 phi = np.linspace(Dphi/2.0,2.0*np.pi-Dphi/2.0,n_phi,endpoint=True)
 #phi = np.linspace(0,2.0*np.pi,n_phi,endpoint=False)
 polar = np.array([np.sin(phi),np.cos(phi)]).T
 fphi= np.sinc(Dphi/2.0/np.pi)
 polar_ave = polar*fphi
 #--------------------

 n_mfp_bulk = len(mfp_bulk) 
 n_mfp = argv.setdefault('n_mfp',100)

 mfp = np.logspace(np.log10(max([min(mfp_bulk),1e-11])),np.log10(max(mfp_bulk)*1.01),n_mfp) 

 n_mfp = len(mfp)

 temp_coeff_p,temp_coeff = np.zeros((2,n_mfp,n_phi))
 kappa_directional_p,kappa_directional = np.zeros((2,n_mfp,n_phi,3)) 
 #suppression_p,suppression = np.zeros((2,n_mfp,n_phi,n_mfp_bulk)) 



 kdp = np.zeros((n_mfp,n_phi,2))
 kd = np.zeros((n_mfp,n_phi,2))
 tcp = np.zeros((n_mfp,n_phi))
 tc = np.zeros((n_mfp,n_phi))
 p = np.arange(n_phi)

 g1 = kappa_bulk/mfp_bulk
 g2 = kappa_bulk/mfp_bulk/mfp_bulk
 

 n_tot = n_mfp_bulk

 block =  n_tot//comm.size
 rr = range(block*comm.rank,n_tot) if comm.rank == comm.size-1 else range(block*comm.rank,block*(comm.rank+1))

 for m in rr:  
       
      (m1,a1,m2,a2) = get_linear_indexes(mfp,mfp_bulk[m])

      tmp = g1[m]*polar_ave
      kdp[m1] += a1*tmp
      kdp[m2] += a2*tmp

      #Temperature
      tmp = g2[m]*Dphi
      tcp[m1] += a1*tmp
      tcp[m2] += a2*tmp

      #Suppression
      #tmp  = dirr[t,:,0]/mfp_bulk[m]
      #sp[m1,:,m] +=  a1 * tmp
      #sp[m2,:,m] +=  a2 * tmp

 comm.Allreduce([kdp,MPI.DOUBLE],[kd,MPI.DOUBLE],op=MPI.SUM)
 comm.Allreduce([tcp,MPI.DOUBLE],[tc,MPI.DOUBLE],op=MPI.SUM)
 #comm.Allreduce([sp,MPI.DOUBLE],[s,MPI.DOUBLE],op=MPI.SUM)

 kd *= 2/n_phi
 #s *= 3*2/4.0/np.pi
 tc /= np.sum(tc)

 #Test for bulk---
 if comm.rank == 0:
  kappa_bulk = np.zeros((2,2))
  for m in range(n_mfp):
   for p in range(n_phi):
      kappa_bulk += mfp[m]*np.outer(kd[m,p],polar_ave[p,:2])

 #Compute RHS average for multiscale modeling---
 rhs_average = mfp_sampled*mfp_sampled/2
 #a = np.zeros((3,3)) 
 #for i in range(len(polar_ave)):
 # a += np.outer(polar_ave[i],polar_ave[i])/2/np.pi*2*np.pi/n_phi
 #rhs_average = mfp_sampled*mfp_sampled
 #rhs_average *= a[0,0]
 #------------------------------------------------
 
 #Final----
 return {'tc':tc,\
         'sigma':kd,\
         'kappa':kappa,\
         'mfp_average':rhs_average*1e18,\
         'VMFP':polar_ave[:,:2],\
         'mfp_sampled':mfp,\
         'phi':phi,\
         'directions': polar  ,\
         'sampling': np.array([n_phi,n_mfp]),\
         'model':np.array([4]),\
         'suppression':np.zeros(1),\
         'kappam':kappa_bulk}





from __future__ import absolute_import
import numpy as np
from scipy.sparse.linalg import splu
from termcolor import colored, cprint 
from .utils import *
from mpi4py import MPI
import scipy.sparse as sp
import time
from scipy.sparse.linalg import lgmres
#import scikits.umfpack as um
import sys
import scipy
from cachetools import cached,LRUCache
from cachetools.keys import hashkey

comm = MPI.COMM_WORLD

cache_assemble = LRUCache(maxsize=1000)
cache_compute_grad_data = LRUCache(maxsize=10000)
cache_get_SU = LRUCache(maxsize=10000)
cache_get_decomposed_directions = LRUCache(maxsize=10000)


def clear_fourier_cache():

   if comm.rank == 0:
    cache_assemble.clear()
    cache_compute_grad_data.clear()
    cache_get_SU.clear()
    cache_get_decomposed_directions.clear()

def fourier_info(data):
          print('                        FOURIER                 ',flush=True)   
          print(colored(' -----------------------------------------------------------','green'),flush=True)
          print(colored('  Iterations:                              ','green') + str(int(data[2])),flush=True)
          print(colored('  Relative error:                          ','green') + '%.1E' % (data[1]),flush=True)
          print(colored('  Fourier Thermal Conductivity [W/m/K]:    ','green') + str(round(data[0],3)),flush=True)
          print(colored(' -----------------------------------------------------------','green'),flush=True)
          print(" ")



@cached(cache=cache_get_SU, key=lambda bundle, k: hashkey(k))
def get_SU(bundle,k):

     A = k*bundle[0] + sp.diags(bundle[1],format='csc')
    
     return splu(A)

def get_key(ll,kappa):


   return (ll,tuple(map(tuple,kappa)))

def unpack(data):

    return data[0],np.array(data[1])


def solve_fourier_single(argv):
   """ Solve Fourier with a single set of kappa """


   data = None

   if comm.rank == 0:

    mesh = argv['geometry']

    mesh['elem_kappa_map'] = get_kappa_map_from_mat(**argv)

    kappa = -1 #it means that it will take the map from mesh

    F,B,scale = assemble(mesh,kappa)

    meta,temp,grad =  solve_convergence(argv,kappa,B,splu(F),scale=scale)

    kappa_map = get_kappa_map(mesh,kappa)

    flux = -np.einsum('cij,cj->ci',kappa_map,grad)

    if argv['verbose']:
     fourier_info(meta)
   
    data = {'meta':meta,'flux':flux,'temperature':temp,'grad':grad}

   return data



def get_kappa(mesh,i,j,ll,kappa):


   if i == j:
    return np.array(kappa[i])
   
   dim = int(mesh['meta'][2])
   normal = mesh['face_normals'][ll,0:dim]

   kappa_i = np.array(kappa[i])
   kappa_j = np.array(kappa[j])

   ki = np.dot(normal,np.dot(kappa_i,normal))
   kj = np.dot(normal,np.dot(kappa_j,normal))
   w  = mesh['interp_weigths'][ll]

   if ll in mesh['interface_sides']:
     dist   = mesh['dists'][ll,0:dim]*1e-9

     h = 0.882e9

     #B = ki*kj/h/np.linalg.norm(dist)
     B = 0

   else:  
     B = 0

   kappa_loc = kj*kappa_i/(ki*(1-w) + kj*w + B)

   return kappa_loc


def get_kappa_tensor(mesh,ll,kappa):


    if np.isscalar(kappa):
       rot = kappa*np.eye(3) 
    elif  kappa.ndim == 3:
       i,j = mesh['side_elem_map_vec'][ll]
       rot = get_kappa(mesh,i,j,ll,kappa)
    else:   
       rot = kappa  

    return rot   


@cached(cache=cache_get_decomposed_directions, key=lambda mesh,kappa_ll:hashkey(kappa_ll))
def get_decomposed_directions(mesh,kappa_ll):
     
    ll,kappa = unpack(kappa_ll)

    kappa_map = get_kappa_tensor(mesh,ll,kappa)

    dim = int(mesh['meta'][2])
    normal = mesh['face_normals'][ll,0:dim]
    dist   = mesh['dists'][ll,0:dim]

    rot = kappa_map[:dim,:dim]
    v_orth = np.dot(normal,np.dot(rot,normal))/np.dot(normal,dist)
    v_non_orth = np.dot(rot,normal) - dist*v_orth
    
    return v_orth,v_non_orth[:dim]


def compute_laplacian(temp,**argv):

  grad_T = compute_grad(temp,**argv).T

  laplacian = np.zeros_like(temp)
  for i,grad in enumerate(grad_T):
     laplacian +=   compute_grad(grad,**argv).T[i]

  return laplacian



@cached(cache=cache_compute_grad_data,key=lambda mesh,dummy:dummy)
def compute_grad_data(mesh,TB):

     #Compute deltas-------------------   
     rr = []
     for i in mesh['side_per_elem']:
       rr.append(i*[0])

     #this is wrong
     for ll in mesh['active_sides'] :
      kc1,kc2 = mesh['side_elem_map_vec'][ll]
      ind1    = list(mesh['elem_side_map_vec'][kc1]).index(ll)
      ind2    = list(mesh['elem_side_map_vec'][kc2]).index(ll)

      delta = 0
      if ll in mesh['periodic_sides']:
         delta = mesh['periodic_side_values'][list(mesh['periodic_sides']).index(ll)]
      else: delta = 0  

      rr[kc1][ind1] = [kc2,kc1, delta] 
      rr[kc2][ind2] = [kc1,kc2,-delta]
 
     return rr


def compute_grad(temp,**argv):

   mesh = argv['geometry'] 

   rr = compute_grad_data(mesh,1)

   diff_temp = [[temp[j[0]]-temp[j[1]]+j[2] for j in f] for f in rr]
   

   #Add boundary
   #if 'TB' in argv.keys():
   # TB  = argv['TB']
   # for n,(eb,sb) in enumerate(zip(*(mesh['eb'],mesh['sb']))):
   #   ind               = list(mesh['elem_side_map_vec'][eb]).index(sb)
   #   diff_temp[eb][ind] = temp[eb]-TB[n]

   gradT = np.array([np.einsum('js,s->j',mesh['weigths'][k,:,:mesh['n_non_boundary_side_per_elem'][k]],np.array(dt)) for k,dt in enumerate(diff_temp)])


   return gradT




def compute_secondary_flux(temp,kappa_map,**argv):

   
   #Compute gradient
   gradT = compute_grad(temp,**argv)
   #--------------

   mesh = argv['geometry'] 
   dim = int(mesh['meta'][2])
   n_elems = len(mesh['elems'])

   #-----------SAVE some time--------------------------------------
   v_non_orth = {}
   for ll in mesh['active_sides']:
      (i,j) = mesh['side_elem_map_vec'][ll]
      if not i==j:

       kappa_loc = get_kappa(mesh,i,j,ll,kappa_map)
       (v_orth,v_non_orth[ll]) = get_decomposed_directions(mesh,get_key(ll,kappa_loc))

       area   = mesh['areas'][ll]   
       normal = mesh['face_normals'][ll,0:dim]
       dist   = mesh['dists'][ll,0:dim]
   #-----------------------------------------------------------------

   C = np.zeros(n_elems)
   for ll in mesh['active_sides']:
      area = mesh['areas'][ll] 
      (i,j) = mesh['side_elem_map_vec'][ll]
      if not i==j:
       w =  mesh['interp_weigths'][ll]
       F_ave = (w*gradT[i] + (1.0-w)*gradT[j])
       #tmp = np.dot(F_ave,cache['v_non_orth'][ll])*area
       tmp = np.dot(F_ave,v_non_orth[ll])*area
       C[i] += tmp
       C[j] -= tmp

   return C/mesh['volumes'],gradT




def compute_diffusive_thermal_conductivity(temp,kappa_map,**argv):
    
   mesh = argv['geometry']  

   dim = int(mesh['meta'][2])
   kappa_eff = 0
   k_non_orth = 0
   area_tot = 0
   for ll in mesh['flux_sides']:

    (i,j) = mesh['side_elem_map_vec'][ll]
    area = mesh['areas'][ll]


    kappa_loc = get_kappa(mesh,i,j,ll,kappa_map)
    (v_orth,v_non_orth) = get_decomposed_directions(mesh,get_key(ll,kappa_loc))

    deltaT = temp[i] - temp[j] - 1

    kappa_eff -= v_orth *  deltaT * area

    if 'grad' in argv.keys():
     gradT = argv['grad']  
     w  =  mesh['interp_weigths'][ll]
     grad_ave = w*gradT[i] + (1.0-w)*gradT[j]
     k_non_orth += np.dot(grad_ave,v_non_orth)/2 * area
    area_tot +=area

   return (kappa_eff+k_non_orth)*mesh['meta'][1]


def solve_convergence(argv,kappa,B,SU,log=False,scale=[]):

    argv.setdefault('max_fourier_error',1e-6) 
    argv.setdefault('max_fourier_iter',10) 

    C  = np.zeros_like(B)

    mesh = argv['geometry']  

    kappa_map = get_kappa_map(mesh,kappa)

    n_elems = len(mesh['elems'])
    dim = int(mesh['meta'][2])
    grad = np.zeros((n_elems,dim))

    
    n_iter = 0;kappa_old = 0;error = 1  
    while error > argv['max_fourier_error'] and \
                  n_iter < argv['max_fourier_iter'] :

        RHS = B + C
    
        if len(scale) > 0: RHS *=scale

        temp = SU.solve(RHS)
        temp = temp - (max(temp)+min(temp))/2.0

        kappa_eff = compute_diffusive_thermal_conductivity(temp,kappa_map,**argv)

        error = abs((kappa_eff - kappa_old)/kappa_eff)
        kappa_old = kappa_eff
        n_iter +=1
        C,grad = compute_secondary_flux(temp,kappa_map,**argv)

    return [kappa_eff,error,n_iter],temp,grad 


def get_kappa_map(mesh,kappa):


    n_elems = len(mesh['elems'])
    dim = int(mesh['meta'][2])

    if np.isscalar(kappa):
        if kappa == -1: 
           kappa = mesh['elem_kappa_map']
        else: 
           print('error')

    elif isinstance(kappa,tuple):
       kappa = np.array(kappa) 
       if kappa.ndim == 2: 
          kappa = np.tile(kappa,(n_elems,1,1))
    elif isinstance(kappa,np.ndarray):
        if kappa.ndim == 2: 
          kappa = np.tile(kappa,(n_elems,1,1))

    kappa = kappa[:,:dim,:dim]
    return kappa


@cached(cache=cache_assemble, key=lambda mesh, kappa: hashkey(kappa))
def assemble(mesh,kappa):

    kappa_map = get_kappa_map(mesh,kappa)

    n_elems = len(mesh['elems'])
    dim = int(mesh['meta'][2])

    iff = [];jff = [];dff = []

    B = np.zeros(n_elems)
    for ll in mesh['active_sides']:

      area = mesh['areas'][ll] 
      (i,j) = mesh['side_elem_map_vec'][ll]
      vi = mesh['volumes'][i]
      vj = mesh['volumes'][j]
      if not i == j:
     
       kappa_loc = get_kappa(mesh,i,j,ll,kappa_map)
      
       (v_orth,dummy) = get_decomposed_directions(mesh,get_key(ll,kappa_loc))

       iff.append(i)
       jff.append(i)
       dff.append(v_orth/vi*area)
       iff.append(i)
       jff.append(j)
       dff.append(-v_orth/vi*area)
       iff.append(j)
       jff.append(j)
       dff.append(v_orth/vj*area)
       iff.append(j)
       jff.append(i)
       dff.append(-v_orth/vj*area)
       if ll in mesh['periodic_sides']:    
        kk = list(mesh['periodic_sides']).index(ll)   
        B[i] += mesh['periodic_side_values'][kk]*v_orth/vi*area
        B[j] -= mesh['periodic_side_values'][kk]*v_orth/vj*area

    F = sp.csc_matrix((np.array(dff),(np.array(iff),np.array(jff))),shape = (n_elems,n_elems))

    #This scale the matrix and fixed zero temperature to a random point
    #scale = fix_instability(F,B)
    scale = np.ones(n_elems)
    return F,B,scale



def solve_fourier(kappa,DeltaT,argv):

    mesh = argv['geometry']
    argv['DeltaT'] = DeltaT
    dim = int(mesh['meta'][2])

    n_serial = len(kappa)
    n_elems = len(mesh['elems'])

    tf = shared_array(np.zeros((n_serial,n_elems)) if comm.rank == 0 else None)
    tfg = shared_array(np.zeros((n_serial,n_elems,dim)) if comm.rank == 0 else None)

    if comm.rank == 0:

     F,B,scale = assemble(mesh,tuple(map(tuple,np.eye(dim))))

     ratio_old = 0
     ratio = np.zeros_like(kappa)
     for m,k in enumerate(kappa):
         
         BB = k*B+DeltaT
         meta,tf[m,:],tfg[m,:,:] = solve_convergence(argv,k*np.eye(dim),BB,get_SU([F,scale],k),scale=scale)
         #meta,tf[m,:],tfg[m,:,:] = solve_convergence(argv,k*np.eye(dim),BB,splu(k*F),scale=scale)
         kappaf = meta[0]
         ratio[m] = kappaf/k

         if m > int(len(kappa)/4):
           error = abs((ratio[m]-ratio[m-1]))/ratio[m]
           if error < 1e-2:
             tf[m:,:]  = tf[m-1]; tfg[m:,:,:] = tfg[m-1]
             ratio[m:] = ratio[m]
             break

     #Regularize---
     if argv.setdefault('experimental_multiscale',False):
      kappa_map = get_kappa_map(mesh,np.eye(dim))
      ratio_0 = compute_diffusive_thermal_conductivity(DeltaT,kappa_map,**argv)
      grad_DeltaT = compute_grad(DeltaT,**argv)
    
      #Regularize gradient
      for m1 in range(len(kappa))[::-1]:
        error = (ratio[m1] - ratio_0)/ratio[m1]
        
        if (ratio[m1] - ratio_0)/ratio[m1] < 1e-1 and abs(ratio[m1]-ratio[m1-1])/ratio[m1] < 1e-1:
            tfg[:m1+1,:,:] = grad_DeltaT
            break

      #Regularize average temperature
      for m1 in range(len(kappa))[::-1]:
        error = np.linalg.norm(tf[m1] - DeltaT)/np.linalg.norm(DeltaT)
        if error < 1e-1 :
            tf[:m1+1,:] = DeltaT
            break
     #-----------------


    comm.Barrier()
    return tf,tfg






import sys
import numpy as np
import scipy
from .utils import *
import time

def rta3D(**argv):


 #Compute directions---------------
 n_phi = int(argv.setdefault('n_phi',48))
 n_theta = int(argv.setdefault('n_phi',24))
 n_mfp = argv.setdefault('n_mfp',50)
 Dphi = 2.0*np.pi/n_phi
 phi = np.linspace(Dphi/2.0,2.0*np.pi-Dphi/2.0,n_phi,endpoint=True)
 Dtheta = np.pi/n_theta
 theta = np.linspace(Dtheta/2,np.pi-Dtheta/2,n_theta,endpoint=True)

 polar = np.array([np.sin(phi),np.cos(phi),np.ones(n_phi)]).T
 azimuthal = np.array([np.sin(theta),np.sin(theta),np.cos(theta)]).T
 direction = np.einsum('lj,kj->lkj',azimuthal,polar)
 direction = direction.reshape((n_theta * n_phi,3))
 #-------------------

 #Import data-----------

 data = load_data(argv.setdefault('filename','rta'))

 
 f = np.divide(np.ones_like(data['tau']), data['tau'], out=np.zeros_like(data['tau']), where=data['tau']!=0)
 kappa = data['kappa']
 #--------------------------

 mfp_bulk = np.einsum('ki,k->ki',data['v'],data['tau'])
 r = np.array([np.linalg.norm(m) for m in mfp_bulk])
 #Eliminate zero MFP modes

 mfp_bulk = mfp_bulk[r > 0]
 C = data['C'][r>0] 
 f = f[r>0] 
 v = data['v'][r>0] 
 tc = C*f
 tc /= tc.sum()
 Jc = np.einsum('k,ki->ki',C,v)

 r = np.array([np.linalg.norm(m) for m in mfp_bulk])

 phi_bulk = np.array([np.arctan2(m[0],m[1]) for m in mfp_bulk])
 phi_bulk[np.where(phi_bulk < 0) ] = 2*np.pi + phi_bulk[np.where(phi_bulk <0)]
 theta_bulk = np.array([np.arccos((m/r[k])[2]) for k,m in enumerate(mfp_bulk)])
 mfp_sampled = np.logspace(-10,np.log10(max(r)*1.01),n_mfp)#min MFP = 1e-1 nm

 #-----------------------
 n_mfp_bulk = len(mfp_bulk) 
 mfp = np.logspace(-1,np.log10(max(r)*1.01),n_mfp)#min MFP = 1e-2 
 n_mfp = len(mfp)
 temp_coeff = np.zeros((n_mfp,n_phi*n_theta))
 kappa_directional = np.zeros((n_mfp,n_phi*n_theta,3)) 

 a = time.time()
 #NEW---
 a1,a2,m1,m2 = fast_interpolation(r,mfp_sampled,bound='extent')
 b1,b2,p1,p2 = fast_interpolation(phi_bulk,phi,bound='periodic')
 c1,c2,t1,t2 = fast_interpolation(theta_bulk,theta,bound='extent')

 index_1 =  t1 * n_phi + p1; u1 = b1*c1
 index_2 =  t1 * n_phi + p2; u2 = b2*c1
 index_3 =  t2 * n_phi + p1; u3 = b1*c2
 index_4 =  t2 * n_phi + p2; u4 = b2*c2

 #Interpolate temperature----
 temp_coeff2 = np.zeros((n_mfp,n_phi*n_theta))
 np.add.at(temp_coeff,(m1,index_1),a1*u1*tc)
 np.add.at(temp_coeff,(m1,index_2),a1*u2*tc)
 np.add.at(temp_coeff,(m2,index_1),a2*u1*tc)
 np.add.at(temp_coeff,(m2,index_2),a2*u2*tc)
 np.add.at(temp_coeff,(m1,index_3),a1*u3*tc)
 np.add.at(temp_coeff,(m1,index_4),a1*u4*tc)
 np.add.at(temp_coeff,(m2,index_3),a2*u3*tc)
 np.add.at(temp_coeff,(m2,index_4),a2*u4*tc)

 #Interpolate flux----
 np.add.at(kappa_directional,(m1, index_1),Jc*(a1*u1)[:,np.newaxis])
 np.add.at(kappa_directional,(m1, index_2),Jc*(a1*u2)[:,np.newaxis])
 np.add.at(kappa_directional,(m2, index_1),Jc*(a2*u1)[:,np.newaxis])
 np.add.at(kappa_directional,(m2, index_2),Jc*(a2*u2)[:,np.newaxis])
 np.add.at(kappa_directional,(m1, index_3),Jc*(a1*u3)[:,np.newaxis])
 np.add.at(kappa_directional,(m1, index_4),Jc*(a1*u4)[:,np.newaxis])
 np.add.at(kappa_directional,(m2, index_3),Jc*(a2*u3)[:,np.newaxis])
 np.add.at(kappa_directional,(m2, index_4),Jc*(a2*u4)[:,np.newaxis])

 
 rhs_average = mfp_sampled*mfp_sampled/3

 return {'tc':temp_coeff,\
         'sigma':kappa_directional,\
         'kappa':kappa,\
         'mfp_average':rhs_average*1e18,\
         'sampling': np.array([n_phi,n_theta,n_mfp]),\
         'VMFP':direction,\
         'mfp_sampled':mfp_sampled,\
         'model':np.array([9])}

   



import sys
import numpy as np
from scipy.sparse.linalg import lsqr 
from scipy.sparse.linalg import inv as sinv
from scipy.linalg import inv as inv
from scipy.sparse import csc_matrix
import scipy
from .utils import *
import time

def energy_conserving(W):


 bottom = np.sum(np.absolute(W))
 r = np.sum(np.absolute(np.einsum('ij->j',W)))
 print('Row check:' + str(r/bottom))
 
 c = np.sum(np.absolute(np.einsum('ij->i',W)))
 print('Column Check:' + str(c/bottom))
 
 nm = np.shape(W)[0]
 delta = np.einsum('ij->j',W)
 b = -2*np.concatenate((delta,delta))
 A = np.zeros((2*nm,2*nm))
 A[0:nm,0:nm] = nm*np.eye(nm)
 A[nm:,nm:] = nm*np.eye(nm)
 A[0:nm,nm:] = 1
 A[nm:,0:nm] = 1
 l = np.linalg.solve(A,b)
 lr = l[:nm]
 lv = l[nm:]
 beta = np.zeros((nm,nm))
 for i in range(nm):
  for j in range(nm):
   beta[i,j] = -(lr[j]+lv[i])/2
 W -= beta

 bottom = np.sum(np.absolute(W))
 r = np.sum(np.absolute(np.einsum('ij->j',W)))
 print('')
 print('   After:')
 print('Row check:' + str(r/bottom))
 c = np.sum(np.absolute(np.einsum('ij->i',W)))
 print('Column Check:' + str(c/bottom))

 return W

def full(**argv):

 filename = argv.setdefault('input_filename','full.npz')

 print(' ')
 print('Importing ' + filename + ' ... ',end= '')
 data = load_data('full')
 print(' done')
 print(' ')

 print('Computing sigmas ... ',end= '')
 v = data['v']
 C = data['C']
 sigma = np.einsum('u,ui->ui',C,v)

 if np.linalg.norm(sigma,axis=0)[2] == 0.0:
     dim = 2
 else:
     dim = 3


 print('... done')
 print(' ')

 print('Simmetrizing scattering matrix A ...',end= '')
 W = data['W']

 #Cut zero velocity modes -----
 tmp = np.linalg.norm(sigma,axis=1) 
 index = (tmp>0).nonzero()[0]
 exclude = (tmp==0).nonzero()[0]
 sigma = sigma[index]
 W = np.delete(W,exclude,0)
 W = np.delete(W,exclude,1)
 tc = C[index]/sum(C[index])
 print(np.shape(W))
 #-----------------------------


 W = 0.5*W + 0.5*W.T
 nm = W.shape[0]

 print('Making W energy conserving ...',end= '')
 W = energy_conserving(W)
 print('... done')
 print(' ')
 print('Computing kappa bulk ...')

 #kappa = np.einsum('ui,uq,qj->ij',sigma,np.linalg.pinv(W),sigma)/data['alpha']
 
 k_xx  = compute_kappa(W*data['alpha'],sigma[:,0])
 k_yy  = compute_kappa(W*data['alpha'],sigma[:,1])
 kappa = np.array([[k_xx,0],[0,k_yy]])


 print(kappa)
 print(' done')
 print(' ')
 #----------

 data = {'tc':tc,'W':W,'sigma':sigma[:,:dim],'kappa':kappa[:dim,:dim],'model':[10],'alpha':data['alpha']}

 return data











import numpy as np
import os
import math
from .full_model import *
from .utils import *
from .mfp2DSym import *
from .mg2DSym import *
from .mfp2D import *
from .mfp3D import *
from .gray2D import *
from .gray2DSym import *
from .rta2DSym import *
from .rta3D import *
from mpi4py import MPI
import shutil

comm = MPI.COMM_WORLD

def Material(**argv):

   model = argv['model']

   #set up database
   source = argv.setdefault('source','local')
   if source == 'database':
    if comm.rank == 0:
        filename = os.path.dirname(os.path.dirname(os.path.realpath(__file__))) + \
                  '/openbte/materials/rta_' +  argv['filename'] + \
                  '_' + str(argv['temperature']) 

        argv['filename'] = filename

   if source == 'unlisted':
    if comm.rank == 0:
      download_file(argv['file_id'],'rta.npz')

   if model == 'rta2DSym':
     if comm.rank == 0:
      data = rta2DSym(**argv)

   elif model == 'full':
     if comm.rank == 0:
      data = full(**argv)

   elif model == 'mfp2DSym':
      data = mfp2DSym(**argv)

   elif model == 'fourier':
      
      kappa = np.eye(3)
      if 'kappa' in argv.keys():
        kappa *= argv['kappa']
      else:  
       kappa[0,0] = argv['kappa_xx']
       kappa[1,1] = argv['kappa_yy']
       kappa[2,2] = argv['kappa_zz']
      data = {'kappa':kappa,'model':[0]}

   elif model == 'mfp2D':
      data = mfp2D(**argv)

   elif model == 'gray2DSym':
      data = gray2DSym(**argv)

   elif model == 'mg2DSym':
     if comm.rank == 0:
      data = mg2DSym(**argv)

   elif model == 'gray2D':
      data = gray2D(**argv)

   #elif model == 'gray3D':
   #   data = mfp3D(**argv)
   
   elif model == 'mfp3D':
      data = mfp3D(**argv)

   elif model == 'rta3D':
     if comm.rank == 0:
      data = rta3D(**argv)

   else:   
      print('No model recognized')
      quit()


   if argv.setdefault('save',True):
     if comm.rank == 0:
         save_data(argv.setdefault('output_filename','material'),data)   

   if comm.rank == 0:
    return data

 






import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)
from pyvtk import *
import numpy as np
import os
from .utils import *
from .viewer import *
from .kappa_mode import *
from .suppression import *
from scipy.spatial import distance
import numpy.ma as ma
from mpi4py import MPI
comm = MPI.COMM_WORLD


def expand_variables(solver,dim):

  #Here we unroll variables for later use--
  variables = {}
  for key,value in solver['variables'].items():
  
     if value['data'].ndim == 1: #scalar
       variables[key] = {'data':value['data'],'units':value['units'],'increment':value['increment']}
       n_elems = len(value['data'])
     elif value['data'].ndim == 2 : #vector 
         variables[key + '(x)'] = {'data':value['data'][:,0],'units':value['units'],'increment':value['increment']}
         variables[key + '(y)'] = {'data':value['data'][:,1],'units':value['units'],'increment':value['increment']}
         if dim == 3: 
             variables[key + '(z)'] = {'data':value['data'][:,2],'units':value['units'],'increment':value['increment']}
         mag = np.array([np.linalg.norm(value) for value in value['data']])
         variables[key + '(mag.)'] = {'data':mag,'units':value['units'],'increment':value['increment']}

  variables['structure'] = {'data':np.zeros(n_elems),'units':'','increment':[0,0,0]}       
  solver['variables'] = variables



def write_vtu_cell(solver,geometry):

   dim = int(geometry['meta'][2])
   output = []
   strc = 'CellData('
   for n,(key, value) in enumerate(solver['variables'].items()):

     if  value['data'].shape[0] == 1:  value['data'] =  value['data'][0] #band-aid solution

     name = key + '[' + value['units'] + ']'
     if value['data'].ndim == 1:
       strc += r'''Scalars(output[''' + str(n) + r'''],name =' ''' + name +  r''' ')'''
       output.append(value['data'])
    
     if value['data'].ndim == 2:

       if dim == 2:
        t = np.zeros_like(value['data'])
        value['data'] = np.concatenate((value['data'],t),axis=1)[:,:3]
       output.append(value['data'])

       strc += r'''Vectors(output[''' + str(n) + r'''],name =' ''' + name +  r''' ')'''

     if n == len(solver['variables'])-1:
      strc += ')'
     else:
      strc += ','

   data = eval(strc)

   if dim == 3:
   
    vtk = VtkData(UnstructuredGrid(geometry['nodes'],tetra=geometry['elems']),data)
   else :
    vtk = VtkData(UnstructuredGrid(geometry['nodes'],triangle=geometry['elems']),data)
   
   vtk.tofile('output','ascii')



def write_vtu(solver,geometry):

   dim = int(geometry['meta'][2])
   output = []
   strc = 'PointData('
   for n,(key, value) in enumerate(solver['variables'].items()):

     if  value['data'].shape[0] == 1:  value['data'] =  value['data'][0] #band-aid solution

     name = key + '[' + value['units'] + ']'
     if value['data'].ndim == 1:
       strc += r'''Scalars(output[''' + str(n) + r'''],name =' ''' + name +  r''' ')'''
       output.append(value['data'])
    
     if value['data'].ndim == 2:

       if dim == 2:
        t = np.zeros_like(value['data'])
        value['data'] = np.concatenate((value['data'],t),axis=1)[:,:3]
       output.append(value['data'])

       strc += r'''Vectors(output[''' + str(n) + r'''],name =' ''' + name +  r''' ')'''

     if n == len(solver['variables'])-1:
      strc += ')'
     else:
      strc += ','

   data = eval(strc)

   if dim == 3:
   
    vtk = VtkData(UnstructuredGrid(geometry['nodes'],tetra=geometry['elems']),data)
   else :
    vtk = VtkData(UnstructuredGrid(geometry['nodes'],triangle=geometry['elems']),data)
   
   vtk.tofile('output','ascii')




def get_surface_nodes(solver,geometry):

     sides = list(geometry['boundary_sides']) + \
                     list(geometry['periodic_sides']) + \
                     list(geometry['inactive_sides'])

     nodes = geometry['sides'][sides].flat


     triangles = np.arange(len(nodes)).reshape((len(sides),3))
     
     for key in solver['variables'].keys():
        solver['variables'][key]['data'] = solver['variables'][key]['data'][nodes]


     geometry['nodes'] = geometry['nodes'][nodes]
     geometry['elems'] = np.array(triangles)




def duplicate_cells(geometry,solver,repeat):

   dim = int(geometry['meta'][2])
   nodes = np.round(geometry['nodes'],4)
   n_nodes = len(nodes)
   #Correction in the case of user's mistake
   if dim == 2:
     repeat[2]  = 1  
   #--------------------  
   size = geometry['size']

   #Create periodic vector
   P = []
   for px in size[0]*np.arange(repeat[0]):
    for py in size[1]*np.arange(repeat[1]):
     for pz in size[2]*np.arange(repeat[2]):
       P.append([px,py,pz])  
   P = np.asarray(P[1:],np.float32)

   #Compute periodic nodes------------------
   pnodes = []
   for s in list(geometry['periodic_sides'])+ list(geometry['inactive_sides']):
     pnodes += list(geometry['sides'][s])
   pnodes = list(np.unique(np.array(pnodes)))
   #--------------------------------------------

   #Repeat Nodes
   #-----------------------------------------------------
   def repeat_cell(axis,n,d,nodes,replace):

    tmp = nodes.copy()
    for x in np.arange(n-1):
     tmp[:,axis] +=  size[axis]

     nodes = np.vstack((nodes,tmp))
    
    return nodes,replace

   replace = {}  
   for i in range(int(geometry['meta'][2])):
     nodes,replace = repeat_cell(i,repeat[i],size[i],nodes,replace)
   #---------------------------------------------------

   #Repeat elements----------
   unit_cell = np.array(geometry['elems']).copy()
  
   elems = geometry['elems'] 
   for i in range(np.prod(repeat)-1):
     elems = np.vstack((elems,unit_cell+(i+1)*(n_nodes)))

   #------------------------

   #duplicate variables---
   for n,(key, value) in enumerate(solver['variables'].items()):  
     unit_cell = value['data'].copy() 
     for nz in range(repeat[2]):
      for ny in range(repeat[1]):
       for nx in range(repeat[0]):
           if nx + ny + nz > 0:  
            inc = value['increment'][0]*nx + value['increment'][1]*ny + value['increment'][2]*nz
            tmp = unit_cell - inc
            if value['data'].ndim == 1:
             value['data'] = np.hstack((value['data'],tmp))
            else: 
             value['data'] = np.vstack((value['data'],tmp))

     solver['variables'][key]['data'] = value['data']

   geometry['elems'] = elems
   geometry['nodes'] = nodes

   #return the new side
   size = [geometry['size'][i] *repeat[i]   for i in range(dim)]

   return size


def get_node_data(solver,geometry):

  dim = int(geometry['meta'][2])
  for key in solver['variables'].keys():
   data = solver['variables'][key]['data']
   #NEW-------------
   conn = np.zeros(len(geometry['nodes']))
   if data.ndim > 1:
       node_data = np.zeros((len(geometry['nodes']),int(dim)))
   else:   
       node_data = np.zeros(len(geometry['nodes']))
    
   #This works only with uniform type of elements 
   elem_flat = np.array(geometry['elems']).flat
   np.add.at(node_data,elem_flat,np.repeat(data,len(geometry['elems'][0]),axis=0))
   np.add.at(conn,elem_flat,np.ones_like(elem_flat))

   if data.ndim > 1:
       np.divide(node_data,conn[:,np.newaxis],out=node_data)
   else: 
       np.divide(node_data,conn,out=node_data)
   #-----------------------
   solver['variables'][key]['data'] = node_data


def Plot(**argv):


  if comm.rank == 0:

   #import data-------------------------------
   geometry     = argv['geometry'] if 'geometry' in argv.keys() else load_data('geometry')
   solver       = argv['solver']   if 'geometry' in argv.keys() else load_data('solver')
   material     = argv['material'] if 'material' in argv.keys() else load_data('material')
   dim = int(geometry['meta'][2])
   model = argv['model']

   if model == 'maps':

     expand_variables(solver,dim) #this is needed to get the component-wise data for plotly

     get_node_data(solver,geometry)
     repeat = argv.setdefault('repeat',[1,1,1])

     if dim == 3:
      get_surface_nodes(solver,geometry)

     size = duplicate_cells(geometry,solver,repeat)

     if argv.setdefault('show',True):
        plot_results(solver,geometry,**argv)

     output ={'nodes'    :geometry['nodes'],\
             'elems'    :geometry['elems'],\
             'variables':solver['variables'],'bulk':solver['kappa_bulk'],'fourier':solver['kappa_fourier'],'size':size,'direction':['x','y','z'][np.argmax(geometry['applied_gradient'])]}

     if 'kappa' in solver.keys():
         output.update({'bte':solver['kappa_bte']}) 

     if argv.setdefault('save_sample',False):
       save_data('sample',output)

     return output


   elif model == 'vtu':

     get_node_data(solver,geometry)

     duplicate_cells(geometry,solver,argv.setdefault('repeat',[1,1,1]))

     write_vtu(solver,geometry)  

   elif model == 'vtu_cell':

     duplicate_cells(geometry,solver,argv.setdefault('repeat',[1,1,1]))
     
     write_vtu_cell(solver,geometry)  


   elif model == 'suppression':

      return plot_suppression(**argv)

   elif model == 'kappa_mode':

      plot_mode_kappa(**argv)









import sys
import numpy as np
import scipy
from .utils import *
import time




def rta_no_interpolation(**argv):


    data = load_data(argv.setdefault('basename','rta'))

    mfp_0 = 1e-10
    mfp_max = 1e-4
    mfp_bulk = np.einsum('ki,k->ki',data['v'],data['tau']) #the minimum MFP is calculated on the real MFP
    I = np.where(np.linalg.norm(mfp_bulk,axis=1) > mfp_0)[0]
    tau = data['tau'][I]
    v = data['v'][I]
    C = data['C'][I]
    f = np.divide(np.ones_like(tau),tau, out=np.zeros_like(tau), where=tau!=0)
    kappam = np.einsum('u,u,u,u->u',tau,C,v[:,0],v[:,0]) 
    Wdiag = C*f
    Jc = np.einsum('k,ki->ki',C,v[:,:2])
    nm = len(Wdiag)

    return  {  'Wdiag':Wdiag,\
               'sigma':Jc,\
               'kappa':data['kappa'],\
               'model':np.array([11])}


def rta2DSym(**argv):

 if not argv.setdefault('interpolation',True):   

     return rta_no_interpolation(**argv)

 #Import data----
 #Get options----
 n_phi = int(argv.setdefault('n_phi',48))
 n_mfp = int(argv.setdefault('n_mfp',50))
 n_theta = 48

 nm = n_phi * n_mfp

 #Create sampled MFPs
 #Polar Angle---------
 Dphi = 2*np.pi/n_phi
 phi = np.linspace(Dphi/2.0,2.0*np.pi-Dphi/2.0,n_phi,endpoint=True)
 #phi = np.linspace(0,2.0*np.pi,n_phi,endpoint=False)
 polar_ave = np.array([np.sin(phi),np.cos(phi),np.zeros(n_phi)]).T

 
 #Import data-----------
 data = load_data(argv.setdefault('filename','rta'))

 #small cut on MFPs
 mfp_0 = 1e-10
 mfp_bulk = np.einsum('ki,k->ki',data['v'],data['tau']) #the minimum MFP is calculated on the real MFP
 I = np.where(np.linalg.norm(mfp_bulk,axis=1) > mfp_0)[0]
 tau = data['tau'][I]
 v = data['v'][I]
 C = data['C'][I]
 f = np.divide(np.ones_like(tau),tau, out=np.zeros_like(tau), where=tau!=0)
 kappam = np.einsum('u,u,u,u->u',tau,C,v[:,0],v[:,0]) 
 tc = C*f
 tc /= tc.sum()
 Jc = np.einsum('k,ki->ki',C,v[:,:2])
 nm = len(tc)
 mfp_bulk = np.einsum('ki,k->ki',v[:,:2],tau) #here we have the 2D one
 r = np.linalg.norm(mfp_bulk,axis=1)
 phi_bulk = np.array([np.arctan2(m[0],m[1]) for m in mfp_bulk])
 phi_bulk[np.where(phi_bulk < 0) ] = 2*np.pi + phi_bulk[np.where(phi_bulk <0)]
 kappa = data['kappa']

 mfp_max = argv.setdefault('mfp_max',np.max(mfp_bulk)*10)
 mfp_min = argv.setdefault('mfp_min',mfp_0)

 mfp_sampled = np.logspace(np.log10(mfp_0)*1.01,np.log10(mfp_max),n_mfp,endpoint=True)#min MFP = 1e-1 nm

 
 #-----------------------
 n_mfp_bulk = len(mfp_bulk) 

 temp_coeff = np.zeros((n_mfp,n_phi))
 kappa_directional = np.zeros((n_mfp,n_phi,2)) 

 lo = 0

 #Interpolation in the MFPs
 r_cut = r[r>mfp_0]
 tc_cut = tc[r>mfp_0]
 Jc_cut = Jc[r>mfp_0]
 phi_cut = phi_bulk[r>mfp_0]

 #Interpolation in the MFPs
 a1,a2,m1,m2 = fast_interpolation(r_cut,mfp_sampled)

 #Interpolation in phi---
 b1,b2,p1,p2 = fast_interpolation(phi_cut,phi,bound='periodic')

 #CHECKED
 np.add.at(temp_coeff,(m1, p1),a1*b1*tc_cut)
 np.add.at(temp_coeff,(m1, p2),a1*b2*tc_cut)
 np.add.at(temp_coeff,(m2, p1),a2*b1*tc_cut)
 np.add.at(temp_coeff,(m2, p2),a2*b2*tc_cut)

 #CHECKED 
 np.add.at(kappa_directional,(m1, p1),Jc_cut*(a1*b1)[:,np.newaxis])
 np.add.at(kappa_directional,(m1, p2),Jc_cut*(a1*b2)[:,np.newaxis])
 np.add.at(kappa_directional,(m2, p1),Jc_cut*(a2*b1)[:,np.newaxis])
 np.add.at(kappa_directional,(m2, p2),Jc_cut*(a2*b2)[:,np.newaxis])

 lo_cut = np.sum(tc[r<mfp_0])
 
 temp_coeff *=(1+lo_cut)

 assert(abs(np.sum(temp_coeff)) -1 < 1e-5)
 #Adjusment for numerical stability temp_coeff /= np.sum(temp_coeff)

 rhs_average = mfp_sampled*mfp_sampled/2

 output =  {'tc':temp_coeff,\
         'sigma':kappa_directional,\
         'kappa':kappa,\
         'mfp_average':rhs_average*1e18,\
         'VMFP':polar_ave[:,:2],\
         'mfp_sampled':mfp_sampled,\
         'sampling': np.array([n_phi,n_theta,n_mfp]),\
         'model':np.array([8])}



 return output
